import pytest
from pytest import approx
from pytest import fixture
import pandas as pd
import greenheart.tools.eco.electrolysis as he_elec
import greenheart.tools.plant_sizing_estimation as gh_sizing
from greenheart.simulation.technologies.hydrogen.electrolysis import PEM_tools
from hopp.utilities import load_yaml
import os
import unittest
import numpy as np
from tests import TEST_ROOT_DIR

input_library_path = TEST_ROOT_DIR / "greenheart" / "test_hydrogen" / "input_files"
project_life_years = 30


TOL = 1e-3

default_config = {
    'project_parameters': 
            {
                'atb_year': 2022,
                'grid_connection': False,
                'ppa_price': 0.025,
                'hybrid_electricity_estimated_cf': 0.492,
                'project_lifetime':30,
            },
    # 'component_sizing': 
    #         {
    #             'hybrid_electricity_estimated_cf': 0.492,
    #             'electrolyzer': 
    #             {
    #                 'resize_for_enduse': False,
    #                 'size_for': 'BOL'
    #                 },
    #         },
    'electrolyzer': 
            {   'sizing': 
                    {
                    'resize_for_enduse': False,
                    'size_for': 'BOL',
                    'hydrogen_dmd': None,
                    },
                'rating': 180,
                'electrolyzer_capex': 700,
                'time_between_replacement': 62320,
                'replacement_cost_percent': 0.15,
                'cost_model': 'singlitico2021',
                'cluster_rating_MW': 40,
                'pem_control_type': 'basic',
                'include_degradation_penalty': True,
                'eol_eff_percent_loss': 13,
                'uptime_hours_until_eol': 77600,  #new
                'turndown_ratio': 0.1, #new
                # 'hydrogen_dmd': None,
            },
    'steel': 
            {   'capacity':{
                    'input_capacity_factor_estimate': 0.9,
                'annual_production_target': 1000000,
            }
            }
    # 'end_use': 
    #         {
    #             'annual_production_target': 1000000,
    #             'estimated_cf': 0.9,
    #             'product': 'steel'
    #         }
}
# NOTE: All test values came from the green steel project repository code.
# Run on Feb 21,2024 for Site 2"

# Test electrolyzer resizing for offgrid and grid
@fixture
def offgrid_resize_config():
    offgrid_resize_config = default_config.copy()
    offgrid_resize_config['project_parameters']['grid_connection'] = False
    offgrid_resize_config['project_parameters']['hybrid_electricity_estimated_cf'] = 0.492
    offgrid_resize_config["electrolyzer"]["sizing"]['resize_for_enduse'] = True
    return offgrid_resize_config

@fixture
def grid_resize_config():
    grid_resize_config = default_config.copy()
    grid_resize_config['project_parameters']['grid_connection'] = True
    grid_resize_config['project_parameters']['hybrid_electricity_estimated_cf'] = 1.0
    grid_resize_config["electrolyzer"]["sizing"]['resize_for_enduse'] = True

    return grid_resize_config

def test_resize_from_steel_grid(grid_resize_config):
    gh_test = gh_sizing.run_resizing_estimation(grid_resize_config)
    assert gh_test["electrolyzer"]["rating"] == 480
    assert gh_test["electrolyzer"]["sizing"]["hydrogen_dmd"] == approx(8366.311517,TOL)

def test_resize_from_steel_offgrid(offgrid_resize_config):
    gh_test = gh_sizing.run_resizing_estimation(offgrid_resize_config)
    assert gh_test["electrolyzer"]["rating"] == 960
    assert gh_test["electrolyzer"]["sizing"]["hydrogen_dmd"] == approx(17004.6982,TOL)

@fixture
def offgrid_physics():
    power_profile_filename = 'GS_offgrid_power_signal.csv'
    offgrid_power_profile_filename = os.path.join(input_library_path,power_profile_filename)
    offgrid_power_profile = pd.read_csv(offgrid_power_profile_filename)

    offgrid_config = default_config.copy()
    offgrid_config['project_parameters']['grid_connection'] = False
    offgrid_config['project_parameters']['hybrid_electricity_estimated_cf'] = 0.492
    offgrid_config["electrolyzer"]["sizing"]['resize_for_enduse'] = False
    offgrid_config['electrolyzer']['rating'] = 960
    
    # offgrid_config["electrolyzer"]["sizing"]["hydrogen_dmd"] = []

    electrolyzer_physics_results = he_elec.run_electrolyzer_physics(offgrid_power_profile, offgrid_config, wind_resource = None, design_scenario='off-grid', show_plots=False, save_plots=False, verbose=False)
    H2_Res_offgrid = electrolyzer_physics_results["H2_Results"]
    electrical_gen_ts_offgrid = electrolyzer_physics_results["power_to_electrolyzer_kw"]
    return [H2_Res_offgrid,electrical_gen_ts_offgrid]

def test_offgrid_electrolyzer_physics(offgrid_physics,subtests):
    H2_Res,power_profile = offgrid_physics
    
    with subtests.test("annual energy kwh input"):
        assert H2_Res['Sim: Total Input Power [kWh]'] == approx(3383382801.267635,TOL)
    
    with subtests.test("BOL rated hydrogen production rate"):
        assert H2_Res['Rated BOL: H2 Production [kg/hr]'] == approx(17579.2991094574,TOL)
    
    with subtests.test("BOL rated power consumption"):
        assert H2_Res['Rated BOL: Power Consumed [kWh]'] == approx(960018.4015366472,TOL)
    
    with subtests.test("BOL rated efficiency"):
        assert H2_Res['Rated BOL: Efficiency [kWh/kg]'] == approx(54.61073251891887, TOL)
    
    with subtests.test("simulation capacity factor"):
        assert H2_Res['Sim: Capacity Factor'] == approx(0.4104064586084918, TOL)
    
    with subtests.test("simulation stack off-cycles"):
        assert H2_Res['Sim: Total Stack Off-Cycles'] == 9483.0
    
    with subtests.test("simulation operation time fraction"):
        assert H2_Res['Sim: Active Time / Sim Time'] == approx(0.8112157534246575,TOL)
    
    with subtests.test("simulation hydrogen production "):
        assert H2_Res['Sim: Total H2 Produced [kg]'] == approx(63200403.136826776,TOL)
    
    with subtests.test("simulation hydrogen warm-up losses"):
        assert H2_Res['Sim: H2 Warm-Up Losses [kg]'] == approx(240892.61322440396,TOL)
    
    with subtests.test("stack life"):
        assert H2_Res['Stack Life [hrs]'] == approx(23633.651537254333,TOL)
    
    with subtests.test("time until replacement"):
        assert H2_Res['Time Until Replacement [hrs]'] == approx(28947.55906846945,TOL)
    
    with subtests.test("life average capacity factor"):
        assert H2_Res['Life: Capacity Factor'] == approx(0.3889240837784226,TOL)
    
    

    
@fixture
def grid_physics():
    grid_config = default_config.copy()
    grid_config['project_parameters']['grid_connection'] = True
    grid_config['project_parameters']['hybrid_electricity_estimated_cf'] = 1.0
    grid_config["electrolyzer"]["sizing"]['resize_for_enduse'] = False
    grid_config['electrolyzer']['rating'] = 480
    grid_config["electrolyzer"]["sizing"]["hydrogen_dmd"] = 8366.311517
    grid_power_profile = []
    electrolyzer_physics_results = he_elec.run_electrolyzer_physics(grid_power_profile, grid_config, wind_resource = None, design_scenario='grid-only', show_plots=False, save_plots=False, verbose=False)
    H2_Res_grid = electrolyzer_physics_results["H2_Results"]
    electrical_gen_ts_grid = electrolyzer_physics_results["power_to_electrolyzer_kw"]
    return [H2_Res_grid,electrical_gen_ts_grid]
    
@fixture
def grid_baseline_power_profile():
    power_profile_filename = 'GS_gridonly_power_signal.csv'
    grid_power_profile_filename = os.path.join(input_library_path,power_profile_filename)
    grid_power_profile = pd.read_csv(grid_power_profile_filename,index_col='Unnamed: 0')
    return grid_power_profile

def test_grid_electrolyzer_physics(grid_physics,grid_baseline_power_profile,subtests):
    H2_Res,power_profile = grid_physics

    with subtests.test("simulation AEP"):
        assert H2_Res['Sim: Total Input Power [kWh]'] == approx(4008524165.6783223,TOL)

    with subtests.test("power profile AEP"):
        assert sum(power_profile) == approx(4008524165.6783223,TOL)

    with subtests.test("power profile trend"):
        assert power_profile[-1] > power_profile[0]
    
    with subtests.test("grid simulation capacity factor"):
        assert H2_Res['Sim: Capacity Factor'] == approx(0.9518373167475502, TOL)

    with subtests.test("grid simulation operation time fraction"):
        assert H2_Res['Sim: Active Time / Sim Time'] == 1.0

    with subtests.test("grid simulation off-cycles"):
        assert H2_Res['Sim: Total Stack Off-Cycles'] == 0

    with subtests.test("grid simulation hydrogen warm-up losses"):
        assert H2_Res['Sim: H2 Warm-Up Losses [kg]'] == 0

    with subtests.test("grid stack life"):
        assert H2_Res['Stack Life [hrs]'] == approx(78049.89571256597,TOL)

    with subtests.test("grid time between replacement"):
        assert H2_Res['Time Until Replacement [hrs]'] == H2_Res['Stack Life [hrs]']

    with subtests.test("grid life average capacity factor"):
        assert H2_Res['Life: Capacity Factor'] == approx(0.9518373167475502,TOL)
    
    with subtests.test("grid power profile start"):
        assert power_profile[0] == approx(grid_baseline_power_profile['combined_hybrid_power_production_hopp'].values[0],TOL)
    
    with subtests.test("grid power profile end"):
        assert power_profile[-1] == approx(grid_baseline_power_profile['combined_hybrid_power_production_hopp'].values[-1],TOL)

def test_electrolyzer_tools(subtests):
    hydrogen_demand = 8366.311517
    cluster_cap_mw = 40
    bol_eff = PEM_tools.get_electrolyzer_BOL_efficiency()
    electrolyzer_capacity_BOL_MW = PEM_tools.size_electrolyzer_for_hydrogen_demand(hydrogen_demand)
    electrolyzer_size_mw = PEM_tools.check_capacity_based_on_clusters(electrolyzer_capacity_BOL_MW,cluster_cap_mw)
    
    with subtests.test("electrolyzer BOL efficiency"):
        assert bol_eff == 54.61

    with subtests.test("electrolyzer size for hydrogen demand"):
        assert np.ceil(electrolyzer_capacity_BOL_MW) == 457

    with subtests.test("electrolyzer size rounded to nearest cluster capacity"):
        assert electrolyzer_size_mw == 480
    


import copy
from typing import Dict, Union, Optional, Tuple
import ProFAST

import pandas as pd
from attrs import define, Factory, field

import os

@define
class Feedstocks:
    """
    Represents the costs and consumption rates of various feedstocks and resources
    used in ammonia production.

    Attributes:
        electricity_cost (float): Cost per MWh of electricity.
        hydrogen_cost (float): Cost per kg of hydrogen.
        cooling_water_cost (float): Cost per gallon of cooling water.
        iron_based_catalyst_cost (float): Cost per kg of iron-based catalyst.
        oxygen_cost (float): Cost per kg of oxygen.
        electricity_consumption (float): Electricity consumption in MWh per kg of
            ammonia production, default is 0.1207 / 1000.
        hydrogen_consumption (float): Hydrogen consumption in kg per kg of ammonia
            production, default is 0.197284403.
        cooling_water_consumption (float): Cooling water consumption in gallons per
            kg of ammonia production, default is 0.049236824.
        iron_based_catalyst_consumption (float): Iron-based catalyst consumption in kg
            per kg of ammonia production, default is 0.000091295354067341.
        oxygen_byproduct (float): Oxygen byproduct in kg per kg of ammonia production,
            default is 0.29405077250145.
    """

    electricity_cost: float
    hydrogen_cost: float
    cooling_water_cost: float
    iron_based_catalyst_cost: float
    oxygen_cost: float
    electricity_consumption: float = 0.1207 / 1000
    hydrogen_consumption = 0.197284403
    cooling_water_consumption = 0.049236824
    iron_based_catalyst_consumption = 0.000091295354067341
    oxygen_byproduct = 0.29405077250145


@define
class AmmoniaCostModelConfig:
    """
    Configuration inputs for the ammonia cost model, including plant capacity and
    feedstock details.

    Attributes:
        plant_capacity_kgpy (float): Annual production capacity of the plant in kg.
        plant_capacity_factor (float): The ratio of actual production to maximum
            possible production over a year.
        feedstocks (Feedstocks): An instance of the `Feedstocks` class detailing the
            costs and consumption rates of resources used in production.
    """

    plant_capacity_kgpy: float
    plant_capacity_factor: float
    feedstocks: Feedstocks


@define
class AmmoniaCosts:
    """
    Base dataclass for calculated costs related to ammonia production, including
    capital expenditures (CapEx) and operating expenditures (OpEx).

    Attributes:
        capex_air_separation_crygenic (float): Capital cost for air separation.
        capex_haber_bosch (float): Capital cost for the Haber-Bosch process.
        capex_boiler (float): Capital cost for boilers.
        capex_cooling_tower (float): Capital cost for cooling towers.
        capex_direct (float): Direct capital costs.
        capex_depreciable_nonequipment (float): Depreciable non-equipment capital costs.
        land_cost (float): Cost of land.
        labor_cost (float): Annual labor cost.
        general_administration_cost (float): Annual general and administrative cost.
        property_tax_insurance (float): Annual property tax and insurance cost.
        maintenance_cost (float): Annual maintenance cost.
        total_fixed_operating_cost (float): Total annual fixed operating cost.
        H2_cost_in_startup_year (float): Hydrogen cost in the startup year.
        energy_cost_in_startup_year (float): Energy cost in the startup year.
        non_energy_cost_in_startup_year (float): Non-energy cost in the startup year.
        variable_cost_in_startup_year (float): Variable cost in the startup year.
        credits_byproduct (float): Credits from byproducts.
    """

    # CapEx
    capex_air_separation_crygenic: float
    capex_haber_bosch: float
    capex_boiler: float
    capex_cooling_tower: float
    capex_direct: float
    capex_depreciable_nonequipment: float
    land_cost: float

    # Fixed OpEx
    labor_cost: float
    general_administration_cost: float
    property_tax_insurance: float
    maintenance_cost: float
    total_fixed_operating_cost: float

    # Feedstock and Byproduct costs
    H2_cost_in_startup_year: float
    energy_cost_in_startup_year: float
    non_energy_cost_in_startup_year: float
    variable_cost_in_startup_year: float
    credits_byproduct: float


@define
class AmmoniaCostModelOutputs(AmmoniaCosts):
    """
    Outputs from the ammonia cost model, extending `AmmoniaCosts` with total capital
    expenditure calculations.

    Attributes:
        capex_total (float): The total capital expenditure for the ammonia plant.
    """

    # CapEx
    capex_total: float


def run_ammonia_model(
    plant_capacity_kgpy: float, plant_capacity_factor: float
) -> float:
    """
    Calculates the annual ammonia production in kilograms based on the plant's
    capacity and its capacity factor.

    Args:
        plant_capacity_kgpy (float): The plant's annual capacity in kilograms per year.
        plant_capacity_factor (float): The capacity factor of the plant, a ratio of
            its actual output over a period of time to its potential output if it
            were possible for it to operate at full capacity continuously over the
            same period.

    Returns:
        float: The calculated annual ammonia production in kilograms per year.
    """
    ammonia_production_kgpy = plant_capacity_kgpy * plant_capacity_factor

    return ammonia_production_kgpy

@define
class AmmoniaCapacityModelConfig:
    """
    Configuration inputs for the ammonia capacity sizing model, including plant capacity and
    feedstock details.

    Attributes:
        hydrogen_amount_kgpy Optional (float): The amount of hydrogen available in kilograms 
            per year to make ammonia.
        desired_ammonia_kgpy Optional (float): The amount of desired ammonia production in
            kilograms per year.
        input_capacity_factor_estimate (float): The estimated ammonia plant capacity factor.
        feedstocks (Feedstocks): An instance of the `Feedstocks` class detailing the
            costs and consumption rates of resources used in production.
    """
    input_capacity_factor_estimate: float
    feedstocks: Feedstocks
    hydrogen_amount_kgpy: Optional[float] = field(default=None)
    desired_ammonia_kgpy: Optional[float] = field(default=None)


    def __attrs_post_init__(self):
        if self.hydrogen_amount_kgpy is None and self.desired_ammonia_kgpy is None:
            raise ValueError("`hydrogen_amount_kgpy` or `desired_ammonia_kgpy` is a required input.")

        if self.hydrogen_amount_kgpy and self.desired_ammonia_kgpy:
            raise ValueError("can only select one input: `hydrogen_amount_kgpy` or `desired_ammonia_kgpy`.")

@define
class AmmoniaCapacityModelOutputs:
    """
    Outputs from the ammonia plant capacity size model.

    Attributes:
        ammonia_plant_capacity_kgpy (float): If amount of hydrogen in kilograms per year is input, 
            the size of the ammonia plant in kilograms per year is output.
        hydrogen_amount_kgpy (float): If amount of ammonia production in kilograms per year is input, 
            the amount of necessary hydrogen feedstock in kilograms per year is output.
    """
    ammonia_plant_capacity_kgpy: float
    hydrogen_amount_kgpy: float

def run_size_ammonia_plant_capacity(config: AmmoniaCapacityModelConfig) -> AmmoniaCapacityModelOutputs:
    """
    Calculates either the annual ammonia production in kilograms based on plant capacity and
    available hydrogen or the amount of required hydrogen based on a desired ammonia production.

    Args:
        config (AmmoniaCapacityModelConfig):
            Configuration object containing all necessary parameters for the capacity sizing,
            including capacity factor estimate and feedstock costs.

    Returns:
        AmmoniaCapacityModelOutputs: An object containing ammonia plant capacity in kilograms
        per year and amount of hydrogen required in kilograms per year.

    """
    if config.hydrogen_amount_kgpy:
        ammonia_plant_capacity_kgpy = (config.hydrogen_amount_kgpy 
            / config.feedstocks.hydrogen_consumption 
            * config.input_capacity_factor_estimate
        )
        hydrogen_amount_kgpy = config.hydrogen_amount_kgpy

    if config.desired_ammonia_kgpy:
        hydrogen_amount_kgpy = (config.desired_ammonia_kgpy
            * config.feedstocks.hydrogen_consumption
            / config.input_capacity_factor_estimate
        )
        ammonia_plant_capacity_kgpy = (config.desired_ammonia_kgpy 
            / config.input_capacity_factor_estimate
        )

    return AmmoniaCapacityModelOutputs(
        ammonia_plant_capacity_kgpy=ammonia_plant_capacity_kgpy,
        hydrogen_amount_kgpy=hydrogen_amount_kgpy
    )

def run_ammonia_cost_model(config: AmmoniaCostModelConfig) -> AmmoniaCostModelOutputs:
    """
    Calculates the various costs associated with ammonia production, including
    capital expenditures (CapEx), operating expenditures (OpEx), and credits from
    byproducts, based on the provided configuration settings.

    Args:
        config (AmmoniaCostModelConfig): Configuration object containing all necessary
            parameters for the cost calculation, including plant capacity, capacity
            factor, and feedstock costs.

    Returns:
        AmmoniaCostModelOutputs: Object containing detailed breakdowns of calculated
            costs, including total capital expenditure, operating costs, and credits
            from byproducts.
    """
    feedstocks = config.feedstocks

    model_year_CEPCI = 596.2  # TODO: what year
    equation_year_CEPCI = 541.7  # TODO: what year

    # scale with respect to a baseline plant (What is this?)
    scaling_ratio = config.plant_capacity_kgpy / (365.0 * 1266638.4)

    # -------------------------------CapEx Costs------------------------------
    scaling_factor_equipment = 0.6

    capex_scale_factor = scaling_ratio**scaling_factor_equipment
    capex_air_separation_crygenic = (
        model_year_CEPCI / equation_year_CEPCI * 22506100 * capex_scale_factor
    )
    capex_haber_bosch = (
        model_year_CEPCI / equation_year_CEPCI * 18642800 * capex_scale_factor
    )
    capex_boiler = model_year_CEPCI / equation_year_CEPCI * 7069100 * capex_scale_factor
    capex_cooling_tower = (
        model_year_CEPCI / equation_year_CEPCI * 4799200 * capex_scale_factor
    )
    capex_direct = (
        capex_air_separation_crygenic
        + capex_haber_bosch
        + capex_boiler
        + capex_cooling_tower
    )
    capex_depreciable_nonequipment = (
        capex_direct * 0.42 + 4112701.84103543 * scaling_ratio
    )
    capex_total = capex_direct + capex_depreciable_nonequipment
    land_cost = capex_depreciable_nonequipment  # TODO: determine if this is the right method or the one in Fixed O&M costs

    # -------------------------------Fixed O&M Costs------------------------------
    scaling_factor_labor = 0.25
    labor_cost = 57 * 50 * 2080 * scaling_ratio**scaling_factor_labor
    general_administration_cost = labor_cost * 0.2
    property_tax_insurance = capex_total * 0.02
    maintenance_cost = capex_direct * 0.005 * scaling_ratio**scaling_factor_equipment
    land_cost = 2500000 * capex_scale_factor
    total_fixed_operating_cost = (
        land_cost
        + labor_cost
        + general_administration_cost
        + property_tax_insurance
        + maintenance_cost
    )

    # -------------------------------Feedstock Costs------------------------------
    H2_cost_in_startup_year = (
        feedstocks.hydrogen_cost
        * feedstocks.hydrogen_consumption
        * config.plant_capacity_kgpy
        * config.plant_capacity_factor
    )
    energy_cost_in_startup_year = (
        feedstocks.electricity_cost
        * feedstocks.electricity_consumption
        * config.plant_capacity_kgpy
        * config.plant_capacity_factor
    )
    non_energy_cost_in_startup_year = (
        (
            (feedstocks.cooling_water_cost * feedstocks.cooling_water_consumption)
            + (
                feedstocks.iron_based_catalyst_cost
                * feedstocks.iron_based_catalyst_consumption
            )
        )
        * config.plant_capacity_kgpy
        * config.plant_capacity_factor
    )
    variable_cost_in_startup_year = (
        energy_cost_in_startup_year + non_energy_cost_in_startup_year
    )
    # -------------------------------Byproduct Costs------------------------------
    credits_byproduct = (
        feedstocks.oxygen_cost
        * feedstocks.oxygen_byproduct
        * config.plant_capacity_kgpy
        * config.plant_capacity_factor
    )

    return AmmoniaCostModelOutputs(
        # Capex
        capex_air_separation_crygenic=capex_air_separation_crygenic,
        capex_haber_bosch=capex_haber_bosch,
        capex_boiler=capex_boiler,
        capex_cooling_tower=capex_cooling_tower,
        capex_direct=capex_direct,
        capex_depreciable_nonequipment=capex_depreciable_nonequipment,
        capex_total=capex_total,
        land_cost=land_cost,
        # Fixed OpEx
        labor_cost=labor_cost,
        general_administration_cost=general_administration_cost,
        property_tax_insurance=property_tax_insurance,
        maintenance_cost=maintenance_cost,
        total_fixed_operating_cost=total_fixed_operating_cost,
        # Feedstock & Byproducts
        H2_cost_in_startup_year=H2_cost_in_startup_year,
        energy_cost_in_startup_year=energy_cost_in_startup_year,
        non_energy_cost_in_startup_year=non_energy_cost_in_startup_year,
        variable_cost_in_startup_year=variable_cost_in_startup_year,
        credits_byproduct=credits_byproduct,
    )


@define
class AmmoniaFinanceModelConfig:
    """
    Configuration for the financial model of an ammonia production plant, including
    operational parameters, cost inputs, and financial assumptions.

    Attributes:
        plant_life (int): Expected operational life of the plant in years.
        plant_capacity_kgpy (float): Annual production capacity of the plant in kilograms.
        plant_capacity_factor (float): The fraction of the year that the plant operates
            at full capacity.
        grid_prices (Dict[str, float]): Electricity prices per kWh, indexed by year.
        feedstocks (Feedstocks): Instance of `Feedstocks` detailing costs and consumption
            rates of inputs.
        costs (Union[AmmoniaCosts, AmmoniaCostModelOutputs]): Pre-calculated capital and
            operating costs for the plant.
        financial_assumptions (Dict[str, float]): Key financial metrics and assumptions
            for the model, such as discount rate and inflation rate. Default is an
            empty dict but should be populated with relevant values.
        install_years (int): Number of years over which the plant is installed and
            ramped up to full production, default is 3 years.
        gen_inflation (float): General inflation rate, default is 0.0.
        save_plots (bool): select whether or not to save output plots
        show_plots (bool): select whether or not to show output plots during run
        output_dir (str): where to store any saved plots or data
        design_scenario_id (int): what design scenario the plots correspond to
    """

    plant_life: int
    plant_capacity_kgpy: float
    plant_capacity_factor: float
    grid_prices: Dict[str, float]
    feedstocks: Feedstocks
    costs: Union[AmmoniaCosts, AmmoniaCostModelOutputs]
    financial_assumptions: Dict[str, float] = Factory(dict)
    install_years: int = 3
    gen_inflation: float = 0.0
    save_plots: bool = False
    show_plots: bool = False
    output_dir: str = "./output/"
    design_scenario_id: int = 0


@define
class AmmoniaFinanceModelOutputs:
    """
    Outputs from the financial model of an ammonia production plant, providing detailed
    financial analysis and projections.

    Attributes:
        sol (dict): Solution to the financial model, containing key performance indicators
            like Net Present Value (NPV), Internal Rate of Return (IRR), and payback
            period.
        summary (dict): Summary of the financial analysis, providing a high-level overview
            of the plant's financial viability.
        price_breakdown (pd.DataFrame): Detailed breakdown of costs contributing to the
            production price of ammonia.
        ammonia_price_breakdown (dict): Breakdown of the ammonia production cost into
            component costs, showing the contribution of each cost element to the
            overall production cost.
    """

    sol: dict
    summary: dict
    price_breakdown: pd.DataFrame


def run_ammonia_finance_model(
    config: AmmoniaFinanceModelConfig,
) -> AmmoniaFinanceModelOutputs:
    """
    Executes the financial analysis for an ammonia production plant based on the
    provided configuration settings. This analysis includes calculating the Net
    Present Value (NPV), Internal Rate of Return (IRR), payback period, and
    providing a detailed cost breakdown for producing ammonia.

    This function leverages the configuration specified in `AmmoniaFinanceModelConfig`,
    including plant operational parameters, grid prices, feedstock costs, pre-calculated
    CapEx and OpEx, and financial assumptions to evaluate the financial performance of
    the ammonia production facility.

    Args:
        config (AmmoniaFinanceModelConfig): Configuration object containing all the
            necessary parameters for the financial analysis, including assumptions
            and pre-calculated cost inputs.

    Returns:
        AmmoniaFinanceModelOutputs: An object containing the results of the financial
            analysis. This includes a solution dictionary with key financial metrics,
            a summary of the financial viability, a price breakdown of ammonia
            production costs, and a detailed breakdown of how each cost component
            contributes to the overall cost of ammonia.
    """
    feedstocks = config.feedstocks
    costs = config.costs

    # Set up ProFAST
    pf = ProFAST.ProFAST("blank")

    # apply all params passed through from config
    for param, val in config.financial_assumptions.items():
        pf.set_params(param, val)

    analysis_start = int(list(config.grid_prices.keys())[0]) - config.install_years

    # Fill these in - can have most of them as 0 also
    pf.set_params(
        "commodity",
        {
            "name": "Ammonia",
            "unit": "kg",
            "initial price": 1000,
            "escalation": config.gen_inflation,
        },
    )
    pf.set_params("capacity", config.plant_capacity_kgpy / 365)  # units/day
    pf.set_params(
        "maintenance",
        {"value": 0, "escalation": config.gen_inflation},
    )
    pf.set_params("analysis start year", analysis_start)
    pf.set_params("operating life", config.plant_life)
    pf.set_params("installation months", 12 * config.install_years)
    pf.set_params(
        "installation cost",
        {
            "value": costs.total_fixed_operating_cost,
            "depr type": "Straight line",
            "depr period": 4,
            "depreciable": False,
        },
    )
    pf.set_params("non depr assets", costs.land_cost)
    pf.set_params(
        "end of proj sale non depr assets",
        costs.land_cost * (1 + config.gen_inflation) ** config.plant_life,
    )
    pf.set_params("demand rampup", 0)
    pf.set_params("long term utilization", config.plant_capacity_factor)
    pf.set_params("credit card fees", 0)
    pf.set_params("sales tax", 0)
    pf.set_params(
        "license and permit", {"value": 00, "escalation": config.gen_inflation}
    )
    pf.set_params("rent", {"value": 0, "escalation": config.gen_inflation})
    pf.set_params("property tax and insurance", 0)
    pf.set_params("admin expense", 0)
    pf.set_params("sell undepreciated cap", True)
    pf.set_params("tax losses monetized", True)
    pf.set_params("general inflation rate", config.gen_inflation)
    pf.set_params("debt type", "Revolving debt")
    pf.set_params("cash onhand", 1)

    # ----------------------------------- Add capital items to ProFAST ----------------
    pf.add_capital_item(
        name="Air Separation by Cryogenic",
        cost=costs.capex_air_separation_crygenic,
        depr_type="MACRS",
        depr_period=7,
        refurb=[0],
    )
    pf.add_capital_item(
        name="Haber Bosch",
        cost=costs.capex_haber_bosch,
        depr_type="MACRS",
        depr_period=7,
        refurb=[0],
    )
    pf.add_capital_item(
        name="Boiler and Steam Turbine",
        cost=costs.capex_boiler,
        depr_type="MACRS",
        depr_period=7,
        refurb=[0],
    )
    pf.add_capital_item(
        name="Cooling Tower",
        cost=costs.capex_cooling_tower,
        depr_type="MACRS",
        depr_period=7,
        refurb=[0],
    )
    pf.add_capital_item(
        name="Depreciable Nonequipment",
        cost=costs.capex_depreciable_nonequipment,
        depr_type="MACRS",
        depr_period=7,
        refurb=[0],
    )

    # -------------------------------------- Add fixed costs--------------------------------
    pf.add_fixed_cost(
        name="Labor Cost",
        usage=1,
        unit="$/year",
        cost=costs.labor_cost,
        escalation=config.gen_inflation,
    )
    pf.add_fixed_cost(
        name="Maintenance Cost",
        usage=1,
        unit="$/year",
        cost=costs.maintenance_cost,
        escalation=config.gen_inflation,
    )
    pf.add_fixed_cost(
        name="Administrative Expense",
        usage=1,
        unit="$/year",
        cost=costs.general_administration_cost,
        escalation=config.gen_inflation,
    )
    pf.add_fixed_cost(
        name="Property tax and insurance",
        usage=1,
        unit="$/year",
        cost=costs.property_tax_insurance,
        escalation=0.0,
    )

    # ---------------------- Add feedstocks, note the various cost options-------------------
    pf.add_feedstock(
        name="Hydrogen",
        usage=feedstocks.hydrogen_consumption,
        unit="kilogram of hydrogen per kilogram of ammonia",
        cost=feedstocks.hydrogen_cost,
        escalation=config.gen_inflation,
    )

    pf.add_feedstock(
        name="Electricity",
        usage=feedstocks.electricity_consumption,
        unit="MWh per kilogram of ammonia",
        cost=config.grid_prices,
        escalation=config.gen_inflation,
    )
    pf.add_feedstock(
        name="Cooling water",
        usage=feedstocks.cooling_water_consumption,
        unit="Gallon per kilogram of ammonia",
        cost=feedstocks.cooling_water_cost,
        escalation=config.gen_inflation,
    )
    pf.add_feedstock(
        name="Iron based catalyst",
        usage=feedstocks.iron_based_catalyst_consumption,
        unit="kilogram of catalyst per kilogram of ammonia",
        cost=feedstocks.iron_based_catalyst_cost,
        escalation=config.gen_inflation,
    )
    pf.add_coproduct(
        name="Oxygen byproduct",
        usage=feedstocks.oxygen_byproduct,
        unit="kilogram of oxygen per kilogram of ammonia",
        cost=feedstocks.oxygen_cost,
        escalation=config.gen_inflation,
    )

    # ------------------------------ Set up outputs ---------------------------

    sol = pf.solve_price()
    summary = pf.get_summary_vals()
    price_breakdown = pf.get_cost_breakdown()

    if config.save_plots or config.show_plots:
        savepaths = [
            config.output_dir + "figures/capex/",
            config.output_dir + "figures/annual_cash_flow/",
            config.output_dir + "figures/lcoa_breakdown/",
            config.output_dir + "data/",
        ]
        for savepath in savepaths:
            if not os.path.exists(savepath):
                os.makedirs(savepath)

        pf.plot_capital_expenses(
            fileout=savepaths[0] + "ammonia_capital_expense_%i.pdf" % (config.design_scenario_id),
            show_plot=config.show_plots,
        )
        pf.plot_cashflow(
            fileout=savepaths[1] + "ammonia_cash_flow_%i.png"
            % (config.design_scenario_id),
            show_plot=config.show_plots,
        )

        pd.DataFrame.from_dict(data=pf.cash_flow_out).to_csv(
            savepaths[3] + "ammonia_cash_flow_%i.csv" % (config.design_scenario_id)
        )

        pf.plot_costs(
            savepaths[2] + "lcoa_%i" % (config.design_scenario_id),
            show_plot=config.show_plots,
        )

    return AmmoniaFinanceModelOutputs(
        sol=sol,
        summary=summary,
        price_breakdown=price_breakdown,
    )

def run_ammonia_full_model(greenheart_config: dict, save_plots=False, show_plots=False, output_dir="./output/", design_scenario_id=0) -> Tuple[AmmoniaCapacityModelOutputs, AmmoniaCostModelOutputs, AmmoniaFinanceModelOutputs]:
    """
    Runs the full ammonia production model, including capacity sizing, cost calculation,

    Args:
        greenheart_config (dict): Configuration settings for the ammonia production model,
            including capacity, costs, and financial assumptions.

    Returns:
        Tuple[AmmoniaCapacityModelOutputs, AmmoniaCostModelOutputs, AmmoniaFinanceModelOutputs]:
            A tuple containing the outputs of the ammonia capacity model, ammonia cost
            model, and ammonia finance model.
    """
    # this is likely to change as we refactor to use config dataclasses, but for now
    # we'll just copy the config and modify it as needed
    config = copy.deepcopy(greenheart_config)

    ammonia_costs = config["ammonia"]["costs"]
    ammonia_capacity = config["ammonia"]["capacity"]
    feedstocks = Feedstocks(**ammonia_costs["feedstocks"])

    # run ammonia capacity model to get ammonia plant size
    capacity_config = AmmoniaCapacityModelConfig(
        feedstocks=feedstocks,
        **ammonia_capacity
    )
    ammonia_capacity = run_size_ammonia_plant_capacity(capacity_config)

    # run ammonia cost model
    ammonia_costs["feedstocks"] = feedstocks
    ammonia_cost_config = AmmoniaCostModelConfig(
        plant_capacity_factor=capacity_config.input_capacity_factor_estimate,
        plant_capacity_kgpy=ammonia_capacity.ammonia_plant_capacity_kgpy,
        **ammonia_costs
    )
    ammonia_cost_config.plant_capacity_kgpy = (
        ammonia_capacity.ammonia_plant_capacity_kgpy
    )
    ammonia_costs = run_ammonia_cost_model(ammonia_cost_config)

    # run ammonia finance model
    ammonia_finance = config["ammonia"]["finances"]
    ammonia_finance["feedstocks"] = feedstocks

    ammonia_finance_config = AmmoniaFinanceModelConfig(
        plant_capacity_kgpy=ammonia_capacity.ammonia_plant_capacity_kgpy,
        plant_capacity_factor=capacity_config.input_capacity_factor_estimate,
        costs=ammonia_costs,
        show_plots=show_plots, 
        save_plots=save_plots,
        output_dir=output_dir,
        design_scenario_id=design_scenario_id,
        **ammonia_finance
    )
    ammonia_finance = run_ammonia_finance_model(ammonia_finance_config)

    return (
        ammonia_capacity,
        ammonia_costs,
        ammonia_finance,
    )


import math
import numpy as np


def hydrogen_storage_capacity(H2_Results, electrolyzer_size_mw, hydrogen_demand_kgphr):
    """Calculate storage capacity based on hydrogen demand and production.

    Args:
        H2_Results (dict): Dictionary including electrolyzer physics results.
        electrolyzer_size_mw (float): Electrolyzer size in MW.
        hydrogen_demand_kgphr (float): Hydrogen demand in kg/hr.

    Returns:
        hydrogen_storage_capacity_kg (float): Hydrogen storage capacity in kilograms.
        hydrogen_storage_duration_hr (float): Hydrogen storage duration in hours using HHV/LHV.
        hydrogen_storage_soc (list): Timeseries of the hydrogen storage state of charge.
    """

    hydrogen_production_kgphr = H2_Results["Hydrogen Hourly Production [kg/hr]"]

    hydrogen_demand_kgphr = max(
        hydrogen_demand_kgphr, np.mean(hydrogen_production_kgphr)
    )  # TODO: potentially add buffer No buffer needed since we are already oversizing

    # TODO: SOC is just an absolute value and is not a percentage. Ideally would calculate as shortfall in future.
    hydrogen_storage_soc = []
    for j in range(len(hydrogen_production_kgphr)):
        if j == 0:
            hydrogen_storage_soc.append(
                hydrogen_production_kgphr[j] - hydrogen_demand_kgphr
            )
        else:
            hydrogen_storage_soc.append(
                hydrogen_storage_soc[j - 1]
                + hydrogen_production_kgphr[j]
                - hydrogen_demand_kgphr
            )

    minimum_soc = np.min(hydrogen_storage_soc)

    #adjust soc so it's not negative.
    if minimum_soc < 0:
        hydrogen_storage_soc = [x + np.abs(minimum_soc) for x in hydrogen_storage_soc]

    hydrogen_storage_capacity_kg = np.max(hydrogen_storage_soc) - np.min(
        hydrogen_storage_soc
    )
    h2_LHV = 119.96  # MJ/kg
    h2_HHV = 141.88  # MJ/kg
    hydrogen_storage_capacity_MWh_LHV = hydrogen_storage_capacity_kg * h2_LHV / 3600
    hydrogen_storage_capacity_MWh_HHV = hydrogen_storage_capacity_kg * h2_HHV / 3600

    # # Get max injection/withdrawal rate
    # hydrogen_injection_withdrawal_rate = []
    # for j in range(len(hydrogen_production_kgphr)):
    #     hydrogen_injection_withdrawal_rate.append(
    #         hydrogen_production_kgphr[j] - hydrogen_demand_kgphr
    #     )
    # max_h2_injection_rate_kgphr = max(hydrogen_injection_withdrawal_rate)

    # # Get storage compressor capacity. TODO: sync compressor calculation here with GreenHEART compressor model
    # compressor_total_capacity_kW = (
    #     max_h2_injection_rate_kgphr / 3600 / 2.0158 * 8641.678424
    # )

    # compressor_max_capacity_kw = 16000
    # n_comps = math.ceil(compressor_total_capacity_kW / compressor_max_capacity_kw)

    # small_positive = 1e-6
    # compressor_avg_capacity_kw = compressor_total_capacity_kW / (
    #     n_comps + small_positive
    # )

    # Get average electrolyzer efficiency
    electrolyzer_average_efficiency_HHV = H2_Results['Sim: Average Efficiency [%-HHV]']

    # Calculate storage durationhyd
    hydrogen_storage_duration_hr = (
        hydrogen_storage_capacity_MWh_LHV
        / electrolyzer_size_mw
        / electrolyzer_average_efficiency_HHV
    )

    return hydrogen_storage_capacity_kg, hydrogen_storage_duration_hr, hydrogen_storage_soc



import numpy as np
import openmdao.api as om

from shapely.geometry import Polygon, Point
from hopp.simulation import HoppInterface
from greenheart.simulation.greenheart_simulation import GreenHeartSimulationConfig, run_simulation

class GreenHeartComponent(om.ExplicitComponent):
    """This class is an OpenMDAO wrapper for running a greenheart simulation
    """
    def initialize(self):
        self.options.declare("config", types=GreenHeartSimulationConfig, recordable=False, desc="GreenHeartSimulationConfig data class instance")
        self.options.declare("outputs_for_finite_difference", default=["lcoh"], types=(type([""])), desc="outputs that should be finite differenced. Must be a list of str.")
        self.options.declare("turbine_x_init", types=(list, type(np.array(0))), desc="Initial turbine easting locations in m")
        self.options.declare("turbine_y_init", types=(list, type(np.array(0))), desc="Initial turbine northing locations in m")
        self.options.declare("verbose", default=False, types=bool, desc="Whether or not to print a bunch of stuff")
        self.options.declare("design_variables", 
                            #  values=["turbine_x", "turbine_y", "pv_capacity_kw", "wind_rating_kw", "electrolyzer_rating_kw", "battery_capacity_kw", "battery_capacity_kwh"], 
                             types=list, 
                             desc="List of design variables that should be included",
                             default=["pv_capacity_kw", "electrolyzer_rating_kw", "battery_capacity_kw", "battery_capacity_kwh"],
                             recordable=False)

    def setup(self):
        ninputs = 0
        hopp_technologies = self.options["config"].hopp_config["technologies"]
        # Add inputs
        if "turbine_x" in self.options["design_variables"]:
           self.add_input("turbine_x", val=self.options["turbine_x_init"], units="m")
           ninputs += len(self.options["turbine_x_init"])
        if "turbine_y" in self.options["design_variables"]:
           self.add_input("turbine_y", val=self.options["turbine_y_init"], units="m")
           ninputs += len(self.options["turbine_y_init"])
        if "wind_rating_kw" in self.options["design_variables"]:
            initial_wind_rating = hopp_technologies["wind"]["num_turbines"]*hopp_technologies["wind"]["turbine_rating_kw"]
            self.add_input("wind_rating_kw", val=initial_wind_rating, units="kW")
            ninputs += 1
        if "pv_capacity_kw" in self.options["design_variables"]:
            self.add_input("pv_capacity_kw", val=hopp_technologies["pv"]["system_capacity_kw"], units="kW")
            ninputs += 1
        if "wave_capacity_kw" in self.options["design_variables"]:
            initial_wave_rating = hopp_technologies["wave"]["device_rating_kw"]*hopp_technologies["wave"]["num_devices"]
            self.add_input("wave_capacity_kw", val=initial_wave_rating, units="kW")
            ninputs += 1
        if "battery_capacity_kw" in self.options["design_variables"]:
            self.add_input("battery_capacity_kw", val=hopp_technologies["battery"]["system_capacity_kw"], units="kW")
            ninputs += 1
        if "battery_capacity_kwh" in self.options["design_variables"]:
            self.add_input("battery_capacity_kwh", val=hopp_technologies["battery"]["system_capacity_kwh"], units="kW*h")
            ninputs += 1
        if ninputs == 0 or "electrolyzer_rating_kw" in self.options["design_variables"]:
            self.add_input("electrolyzer_rating_kw", val=self.options["config"].greenheart_config["electrolyzer"]["rating"]*1E3, units="kW")
            ninputs += 1

        # add outputs
        self.add_output("lcoe", units="USD/(kW*h)", val=0.0, desc="levelized cost of energy")
        self.add_output("lcoh", units="USD/kg", val=0.0, desc="levelized cost of hydrogen")
        if "steel" in self.options["config"].greenheart_config.keys():
            self.add_output("lcos", units="USD/t", val=0.0, desc="levelized cost of steel")
        if "ammonia" in self.options["config"].greenheart_config.keys():
            self.add_output("lcoa", units="USD/kg", val=0.0, desc="levelized cost of ammonia")

        if self.options["config"].greenheart_config["opt_options"]["constraints"]["pv_to_platform_area_ratio"]["flag"]:

            self.add_output("pv_area", units="m*m", val=0.0, desc="offshore pv array area")
            self.add_output("platform_area", units="m*m", val=0.0, desc="offshore platform area")

            self.options["outputs_for_finite_difference"].append("pv_area")
            self.options["outputs_for_finite_difference"].append("platform_area")

    def compute(self, inputs, outputs):

        if self.options["verbose"]:
            print("reinitialize")

        config = self.options["config"]

        if any(x in ["wind_rating_kw", "pv_capacity_kw", "battery_capacity_kw", "battery_capacity_kwh"] for x in inputs):
            if "wind_rating_kw" in inputs:
                raise(NotImplementedError("wind_rating_kw has not be fully implemented as a design variable"))
            if "pv_capacity_kw" in inputs:
                config.hopp_config["technologies"]["pv"]["system_capacity_kw"] = float(inputs["pv_capacity_kw"])
            if "battery_capacity_kw" in inputs:
                config.hopp_config["technologies"]["battery"]["system_capacity_kw"] = float(inputs["battery_capacity_kw"])
            if "battery_capacity_kwh" in inputs:
                config.hopp_config["technologies"]["battery"]["system_capacity_kwh"] = float(inputs["battery_capacity_kwh"])
        if "electrolyzer_rating_kw" in inputs:
            config.greenheart_config["electrolyzer"]["rating"] = float(inputs["electrolyzer_rating_kw"])*1E-3

        if config.output_level == 0:
            run_simulation(config)
        elif config.output_level == 1:
            lcoh = run_simulation(config)
        elif config.output_level == 2:
            lcoh, lcoe, capex_breakdown, opex_breakdown_annual, pf_lcoh, electrolyzer_physics_results = run_simulation(config)
        elif config.output_level == 3:
            lcoh, lcoe, capex_breakdown, opex_breakdown_annual, pf_lcoh, electrolyzer_physics_results, pf_lcoe, power_breakdown = run_simulation(config)
        elif config.output_level == 4:
            lcoe, lcoh, lcoh_grid_only = run_simulation(config)
        elif config.output_level == 5:
            lcoe, lcoh, lcoh_grid_only, hopp_interface = run_simulation(config)
        elif config.output_level == 6:
            hopp_results, electrolyzer_physics_results, remaining_power_profile = run_simulation(config)
        elif config.output_level == 7:
            lcoe, lcoh, steel_finance, ammonia_finance = run_simulation(config)
        elif config.output_level == 8:
            greenheart_output = run_simulation(config)
            lcoe = greenheart_output.lcoe
            lcoh = greenheart_output.lcoh
            steel_finance = greenheart_output.steel_finance
            ammonia_finance = greenheart_output.ammonia_finance
            pv_area = greenheart_output.hopp_results['hybrid_plant'].pv.footprint_area
            platform_area = greenheart_output.platform_results["toparea_m2"]

        outputs["lcoe"] = lcoe
        outputs["lcoh"] = lcoh
        
        if "steel" in self.options["config"].greenheart_config.keys():
            outputs["lcos"] = steel_finance.sol.get("price")
        if "ammonia" in self.options["config"].greenheart_config.keys():
            outputs["lcoa"] = ammonia_finance.sol.get("price")

        if self.options["config"].greenheart_config["opt_options"]["constraints"]["pv_to_platform_area_ratio"]["flag"]:
            outputs["pv_area"] = pv_area
            outputs["platform_area"] = platform_area

    def setup_partials(self):
        self.declare_partials(self.options["outputs_for_finite_difference"], '*', method='fd', form="forward")
        
class HOPPComponent(om.ExplicitComponent):
    """This class is an OpenMDAO wrapper for running a HOPP simulation
    """

    def initialize(self):
        self.options.declare("hi", types=HoppInterface, recordable=False, desc="HOPP Interface class instance")
        self.options.declare("turbine_x_init", types=(list, type(np.array(0))), desc="Initial turbine easting locations in m")
        self.options.declare("turbine_y_init", types=(list, type(np.array(0))), desc="Initial turbine northing locations in m")
        self.options.declare("verbose", default=False, types=bool, desc="Whether or not to print a bunch of stuff")
        self.options.declare("design_variables", 
                            #  values=["turbine_x", "turbine_y", "pv_capacity_kw", "wind_rating_kw", "electrolyzer_rating_kw", "battery_capacity_kw", "battery_capacity_kwh"], 
                             types=list, 
                             desc="List of design variables that should be included",
                             default=["turbine_x", "turbine_y"],
                             recordable=False)

    def setup(self):

        ninputs = 0
        if "turbine_x" in self.options["design_variables"]:
            self.add_input("turbine_x", val=self.options["turbine_x_init"], units="m")
            ninputs += len(self.options["turbine_x_init"])
        if "turbine_y" in self.options["design_variables"]:
            self.add_input("turbine_y", val=self.options["turbine_y_init"], units="m")
            ninputs += len(self.options["turbine_y_init"])
        if "wind_rating_kw" in self.options["design_variables"]:
            self.add_input("wind_rating_kw", val=150000, units="kW")
            ninputs += 1
        if "pv_capacity_kw" in self.options["design_variables"]:
            self.add_input("pv_capacity_kw", val=15000, units="kW")
            ninputs += 1
        if "battery_capacity_kw" in self.options["design_variables"]:
            self.add_input("battery_capacity_kw", val=15000, units="kW")
            ninputs += 1
        if "battery_capacity_kwh" in self.options["design_variables"]:
            self.add_input("battery_capacity_kwh", val=15000, units="kW*h")
            ninputs += 1
        if ninputs == 0:
            self.add_input("turbine_x", val=self.options["turbine_x_init"], units="m")
            
        technologies = self.options["hi"].configuration["technologies"]

        if "pv" in technologies.keys():
            self.add_output("pv_capex", val=0.0, units="USD")
            # self.add_output("pv_opex", val=0.0, units="USD")
        if "wind" in technologies.keys():
            self.add_output("wind_capex", val=0.0, units="USD")
            # self.add_output("wind_opex", val=0.0, units="USD")
        if "battery" in technologies.keys():
            self.add_output("battery_capex", val=0.0, units="USD")
            # self.add_output("battery_opex", val=0.0, units="USD")

        self.add_output("hybrid_electrical_generation_capex", units="USD")
        self.add_output("hybrid_electrical_generation_opex", units="USD")
        self.add_output("aep", units="kW*h")
        self.add_output("lcoe_real", units="USD/(MW*h)")
        self.add_output("power_signal", units="kW", val=np.zeros(8760))

    def compute(self, inputs, outputs):

        hi = self.options["hi"]
        technologies = hi.configuration["technologies"]
        
        if any(x in ["wind_rating_kw", "pv_capacity_kw", "battery_capacity_kw", "battery_capacity_kwh"] for x in inputs):
            if "wind_rating_kw" in inputs:
                raise(NotImplementedError("wind_rating_kw has not be fully implemented as a design variable"))
            if "pv_capacity_kw" in inputs:
                technologies["pv"]["system_capacity_kw"] = float(inputs["pv_capacity_kw"])
            if "battery_capacity_kw" in inputs:
                technologies["battery"]["system_capacity_kw"] = float(inputs["battery_capacity_kw"])
            if "battery_capacity_kwh" in inputs:
                technologies["battery"]["system_capacity_kwh"] = float(inputs["battery_capacity_kwh"])

            configuration = hi.configuration
            configuration["technologies"] = technologies
            hi.reinitialize(configuration)
        
        if ("turbine_x" in inputs) or ("turbine_y" in inputs):
            if "turbine_x" not in inputs:
                hi.system.wind._system_model.fi.set(layout_y=inputs["turbine_y"])
            elif "turbine_y" not in inputs:
                hi.system.wind._system_model.fi.set(layout_x=inputs["turbine_x"])
            else:
                hi.system.wind._system_model.fi.set(
                    layout_x=inputs["turbine_x"], layout_y=inputs["turbine_y"]
                )
                
        # run simulation
        hi.simulate(25)

        if self.options["verbose"]:
            print(f"obj: {hi.system.annual_energies.hybrid}")

        # get results
        if "pv" in technologies.keys():
            outputs["pv_capex"] = hi.system.pv.total_installed_cost
            # outputs["pv_opex"] = hi.system.pv.om_total_expense[1]
        if "wind" in technologies.keys():
            outputs["wind_capex"] = hi.system.wind.total_installed_cost
            # outputs["wind_opex"] = hi.system.wind.om_total_expense[1]
        if "battery" in technologies.keys():
            outputs["battery_capex"] = hi.system.battery.total_installed_cost
            # outputs["battery_opex"] = hi.system.battery.om_total_expense[1]

        outputs["hybrid_electrical_generation_capex"] = hi.system.cost_installed["hybrid"]
        # outputs["hybrid_electrical_generation_opex"] = hi.system.om_total_expenses[1]

        outputs["aep"] = hi.system.annual_energies.hybrid
        outputs["lcoe_real"] = hi.system.lcoe_real.hybrid/100.0 # convert from cents/kWh to USD/kWh
        outputs["power_signal"] = hi.system.grid.generation_profile[0:8760]

    def setup_partials(self):
        self.declare_partials(['lcoe_real', 'power_signal'], '*', method='fd', form="forward")

class TurbineDistanceComponent(om.ExplicitComponent):
    """This class is an OpenMDAO component for placing a constraint on wind turbine distances
    """

    def initialize(self):
        self.options.declare("turbine_x_init")
        self.options.declare("turbine_y_init")

    def setup(self):
        n_turbines = len(self.options["turbine_x_init"])
        n_distances = int(n_turbines*(n_turbines-1)/2)
        self.add_input("turbine_x", val=self.options["turbine_x_init"], units="m")
        self.add_input("turbine_y", val=self.options["turbine_y_init"], units="m")
        self.add_output("spacing_vec", val=np.zeros(n_distances), units="m")

    def compute(self, inputs, outputs):

        n_turbines = len(inputs["turbine_x"])
        spacing_vec = np.zeros(int(n_turbines*(n_turbines-1)/2))
        k = 0
        for i in range(len(inputs["turbine_x"])):
            for j in range(i+1, len(inputs["turbine_y"])):  # only look at pairs once
                spacing_vec[k] = np.linalg.norm([(inputs["turbine_x"][i]-inputs["turbine_x"][j]), (inputs["turbine_y"][i]-inputs["turbine_y"][j])])
                k += 1
        outputs["spacing_vec"] = spacing_vec
        
    def setup_partials(self):
        self.declare_partials('*', '*', method='fd', form='forward')

class BoundaryDistanceComponent(om.ExplicitComponent):
    """This class is an OpenMDAO component for placing a boundary constraint on wind turbine locations
    """
    def initialize(self):
        self.options.declare("hopp_interface", recordable=False)
        self.options.declare("turbine_x_init")
        self.options.declare("turbine_y_init")

    def setup(self):
        self.n_turbines = len(self.options["turbine_x_init"])
        self.n_distances = self.n_turbines
        self.add_input("turbine_x", val=self.options["turbine_x_init"], units="m")
        self.add_input("turbine_y", val=self.options["turbine_y_init"], units="m")
        self.add_output("boundary_distance_vec", val=np.zeros(self.n_distances), units="m")

    def compute(self, inputs, outputs):

        # get hopp interface
        hi = self.options["hopp_interface"]

        # get polygon for boundary
        boundary_polygon = Polygon(hi.system.site.vertices) 
        
        # check if turbines are inside polygon and get distance
        for i in range(0, self.n_distances):
            point = Point(inputs["turbine_x"][i], inputs["turbine_y"][i])
            outputs["boundary_distance_vec"][i] = boundary_polygon.exterior.distance(point)
            if not boundary_polygon.contains(point):
                outputs["boundary_distance_vec"][i] *= -1

    def setup_partials(self):
        self.declare_partials('*', '*', method='fd', form='forward')



"""
This file is based on the WISDEM file of the same name: https://github.com/WISDEM/WISDEM
"""

import os
import pickle

import numpy as np
import pandas as pd
import scipy.io as sio


def save_data(fname, prob, npz_file=True, mat_file=True, xls_file=True):
    # Remove file extension
    froot = os.path.splitext(fname)[0]

    # Get all OpenMDAO inputs and outputs into a dictionary
    var_dict = prob.model.list_inputs(prom_name=True, units=True, desc=True, out_stream=None)
    out_dict = prob.model.list_outputs(prom_name=True, units=True, desc=True, out_stream=None)
    var_dict.extend(out_dict)

    # Pickle the full archive so that we can load it back in if we need
    with open(froot + ".pkl", "wb") as f:
        pickle.dump(var_dict, f)

    # Reduce to variables we can save for matlab or python
    if npz_file or mat_file:
        array_dict = {}
        for k in range(len(var_dict)):
            unit_str = var_dict[k][1]["units"]
            if unit_str is None or unit_str == "Unavailable":
                unit_str = ""
            elif len(unit_str) > 0:
                unit_str = "_" + unit_str

            iname = var_dict[k][1]["prom_name"] + unit_str
            value = var_dict[k][1]["val"]

            if iname in array_dict:
                continue

            if type(value) in [type(np.array([])), type(0.0), type(0), np.float64, np.int64]:
                array_dict[iname] = value
            elif type(value) == type(True):
                array_dict[iname] = np.bool_(value)
            elif type(value) == type(""):
                array_dict[iname] = np.str_(value)
            elif type(value) == type([]):
                temp_val = np.empty(len(value), dtype=object)
                temp_val[:] = value[:]
                array_dict[iname] = temp_val
            # else:
            #    print(var_dict[k])

    # Save to numpy compatible
    if npz_file:
        kwargs = {key: array_dict[key] for key in array_dict.keys()}
        np.savez_compressed(froot + ".npz", **kwargs)

    # Save to matlab compatible
    if mat_file:
        sio.savemat(froot + ".mat", array_dict, long_field_names=True)

    if xls_file:
        data = {}
        data["variables"] = []
        data["units"] = []
        data["values"] = []
        data["description"] = []
        for k in range(len(var_dict)):
            unit_str = var_dict[k][1]["units"]
            if unit_str is None:
                unit_str = ""

            iname = var_dict[k][1]["prom_name"]
            if iname in data["variables"]:
                continue

            data["variables"].append(iname)
            data["units"].append(unit_str)
            data["values"].append(var_dict[k][1]["val"])
            data["description"].append(var_dict[k][1]["desc"])
        df = pd.DataFrame(data)
        df.to_excel(froot + ".xlsx", index=False)
        df.to_csv(froot + ".csv", index=False)


def load_data(fname, prob):
    # Remove file extension
    froot = os.path.splitext(fname)[0]

    # Load in the pickled data
    with open(froot + ".pkl", "rb") as f:
        var_dict = pickle.load(f)

    # Store into Problem object
    for k in range(len(var_dict)):
        iname = var_dict[k][0]
        iname2 = var_dict[k][1]["prom_name"]
        value = var_dict[k][1]["val"]
        try:
            prob.set_val(iname, value)
        except:
            pass
        try:
            prob.set_val(iname2, value)
        except:
            pass

    return prob



"""
This file is based on the WISDEM file 'runWISDEM.py`: https://github.com/WISDEM/WISDEM
"""

import os
import sys
import logging
import warnings

import numpy as np
import openmdao.api as om


from greenheart.simulation.greenheart_simulation import GreenHeartSimulationConfig, setup_greenheart_simulation
from greenheart.tools.optimization.gc_PoseOptimization import PoseOptimization
from greenheart.tools.optimization.openmdao import GreenHeartComponent
from greenheart.tools.optimization.mpi_tools import MPI, map_comm_heirarchical
from greenheart.tools.optimization import fileIO

def run_greenheart(config:GreenHeartSimulationConfig, overridden_values=None, run_only=False):
    """This functions sets up and runs greenheart. It can be used for analysis runs, optimizations, design of experiments, or step size studies

    Args:
        config (GreenHeartSimulationConfig): data structure class containing all simulation options
        overridden_values (_type_, optional): data values from `config` may be overridden using this input at call time. Defaults to None.
        run_only (bool, optional): if True, not optimization or design of experiments will be run. Defaults to False.

    Returns:
        prob: an openmdao problem instance
        config: see Args
    """
    # Initialize openmdao problem. If running with multiple processors in MPI, use parallel finite differencing equal to the number of cores used.
    # Otherwise, initialize the GreenHEART system normally. Get the rank number for parallelization. We only print output files using the root processor.
    myopt = PoseOptimization(config)

    if MPI:
        n_DV = myopt.get_number_design_variables()

        # Extract the number of cores available
        max_cores = MPI.COMM_WORLD.Get_size()

        if max_cores > n_DV and not config.greenheart_config["opt_options"]["driver"]["design_of_experiments"]["flag"]:
            raise ValueError(
                "ERROR: please reduce the number of cores, currently set to "
                + str(max_cores)
                + ", to the number of finite differences "
                + str(n_DV)
                + ", which is equal to the number of design variables DV for forward differencing"
                + " and DV times 2 for central differencing,"
                + " or the parallelization logic will not work"
            )
        
        if config.greenheart_config["opt_options"]["driver"]["design_of_experiments"]["flag"]:
            n_FD = max_cores
            
        else:
            # Define the color map for the parallelization, determining the maximum number of parallel finite difference (FD) evaluations based on the number of design variables (DV).
            n_FD = min([max_cores, n_DV])

            # Define the color map for the cores
            n_FD = max([n_FD, 1])

        comm_map_down, comm_map_up, color_map = map_comm_heirarchical(n_FD, 1)
        rank = MPI.COMM_WORLD.Get_rank()
        color_i = color_map[rank]
        comm_i = MPI.COMM_WORLD.Split(color_i, 1)
    else:
        color_i = 0
        rank = 0

    folder_output = config.output_dir

    if "opt_options" in config.greenheart_config.keys():
        design_variables = list()
        for key in config.greenheart_config["opt_options"]["design_variables"].keys():
            if config.greenheart_config["opt_options"]["design_variables"][key]["flag"]:
                design_variables.append(key)
    else:
        design_variables = []

    if rank == 0 and "opt_options" in config.greenheart_config.keys():
        os.makedirs(folder_output, exist_ok=True)

        # create logger
        logger = logging.getLogger("greenheart")
        logger.setLevel(logging.INFO)

        # create handlers
        ht = logging.StreamHandler()
        ht.setLevel(logging.WARNING)

        flog = os.path.join(folder_output, config.greenheart_config["opt_options"]["general"]["fname_output"] + ".log")
        hf = logging.FileHandler(flog, mode="w")
        hf.setLevel(logging.INFO)

        # create formatters
        formatter_t = logging.Formatter("%(module)s:%(funcName)s:%(lineno)d %(levelname)s:%(message)s")
        formatter_f = logging.Formatter(
            "P%(process)d %(asctime)s %(module)s:%(funcName)s:%(lineno)d %(levelname)s:%(message)s"
        )

        # add formatter to handlers
        ht.setFormatter(formatter_t)
        hf.setFormatter(formatter_f)

        # add handlers to logger
        logger.addHandler(ht)
        logger.addHandler(hf)
        logger.info("Started")

    if color_i == 0:  # the top layer of cores enters
        if MPI:
            # Parallel settings for OpenMDAO
            prob = om.Problem(model=om.Group(num_par_fd=n_FD), comm=comm_i, reports=False)
            
        else:
            # Sequential finite differencing
            prob = om.Problem(model=om.Group(), reports=False)

        prob.model.add_subsystem(
                'greenheart', GreenHeartComponent(config=config, design_variables=design_variables),  
                promotes=["*"])
        
        # If at least one of the design variables is active, setup an optimization
        if not run_only and config.greenheart_config["opt_options"]["opt_flag"]:
            config, hi, _ = setup_greenheart_simulation(config)
            prob = myopt.set_driver(prob)
            prob = myopt.set_objective(prob)
            prob = myopt.set_design_variables(prob, config, hi)
            prob = myopt.set_constraints(prob, hi)

        if config.greenheart_config["opt_options"]["recorder"]["flag"]:
            prob = myopt.set_recorders(prob)

        # Setup openmdao problem
        prob.setup()

        # If the user provides values in this dict, they overwrite
        # whatever values have been set by the yaml files.
        # This is useful for performing black-box wrapped optimization without
        # needing to modify the yaml files.
        if overridden_values is not None:
            for key in overridden_values:
                prob[key] = overridden_values[key]

        # Place the last design variables from a previous run into the problem.
        # This needs to occur after the above setup() and yaml2openmdao() calls
        # so these values are correctly placed in the problem.

        if not run_only:
            prob = myopt.set_restart(prob)

            if "check_totals" in config.greenheart_config["opt_options"]["driver"]:
                if config.greenheart_config["opt_options"]["driver"]["check_totals"]:
                    prob.run_model()
                    totals = prob.compute_totals()

            if "check_partials" in config.greenheart_config["opt_options"]["driver"]:
                if config.greenheart_config["opt_options"]["driver"]["check_partials"]:
                    prob.run_model()
                    checks = prob.check_partials(compact_print=True)

            sys.stdout.flush()
            
            if config.greenheart_config["opt_options"]["driver"]["step_size_study"]["flag"]:
                prob.run_model()
                study_options = config.greenheart_config["opt_options"]["driver"]["step_size_study"]
                step_sizes = study_options["step_sizes"]
                all_derivs = {}
                for idx, step_size in enumerate(step_sizes):
                    prob.model.approx_totals(method="fd", step=step_size, form=study_options["form"])

                    if study_options["of"]:
                        of = study_options["of"]
                    else:
                        of = None

                    if study_options["wrt"]:
                        wrt = study_options["wrt"]
                    else:
                        wrt = None

                    derivs = prob.compute_totals(of=of, wrt=wrt, driver_scaling=study_options["driver_scaling"])
                    all_derivs[idx] = derivs
                    all_derivs[idx]["step_size"] = step_size
                np.save("total_derivs.npy", all_derivs)

            # Run openmdao problem
            elif config.greenheart_config["opt_options"]["opt_flag"]:
                prob.run_driver()
        else:
            prob.run_model()
        if config.greenheart_config["opt_options"]["recorder"]["flag"]:
            prob.record("final_state")

        if (not MPI) or (MPI and rank == 0):
            # Save data coming from openmdao to an output yaml file
            froot_out = os.path.join(folder_output, config.greenheart_config["opt_options"]["general"]["fname_output"])

            # Save data to numpy and matlab arrays
            fileIO.save_data(froot_out, prob)

    if rank == 0:
        return prob, config
    else:
        return [], []




from pytest import raises
from greenheart.tools.eco.utilities import ceildiv, visualize_plant


def test_visualize_plant(subtests):
    
    with subtests.test("'visualize_plant()' only works with the 'floris' wind model"):
        hopp_config ={"technologies": {"wind": {"model_name": "pysam"}}}
        with raises(NotImplementedError, match="only works with the 'floris' wind model"):
            visualize_plant(hopp_config, None, None, None, None, None, None, None, None, None, None, None)

def test_ceildiv(subtests):
    
    with subtests.test("ceildiv"):
        a = 8
        b = 3
        
        assert ceildiv(a, b) == 3
    
    with subtests.test("ceildiv with one negative value"):
        a = 8
        b = -3
        
        assert ceildiv(a, b) == -2

    with subtests.test("ceildiv with two negative values"):
        a = -8
        b = -3
        
        assert ceildiv(a, b) == 3


import os
import os.path
from pathlib import Path
import yaml
import copy

import numpy as np
import numpy_financial as npf
import pandas as pd

import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.ticker as ticker

import ORBIT as orbit

from hopp.simulation.technologies.resource.wind_resource import WindResource

from hopp.simulation import HoppInterface

from hopp.utilities import load_yaml

from hopp.tools.dispatch import plot_tools

from .finance import adjust_orbit_costs

"""
This function returns the ceiling of a/b (rounded to the nearest greater integer). 
The function was copied from https://stackoverflow.com/a/17511341/5128616
"""
def ceildiv(a, b):
    return -(a // -b)

def convert_relative_to_absolute_path(config_filepath, resource_filepath):
    if resource_filepath == "":
        return ""
    else:
        abs_config_filepath = Path(config_filepath).absolute().parent
        return os.path.join(abs_config_filepath, resource_filepath)

# Function to load inputs
def get_inputs(
    filename_hopp_config,
    filename_greenheart_config,
    filename_orbit_config,
    filename_turbine_config,
    filename_floris_config=None,
    verbose=False,
    show_plots=False,
    save_plots=False,
):

    ############### load turbine inputs from yaml

    # load turbine inputs
    turbine_config = load_yaml(filename_turbine_config)

    # load hopp inputs
    hopp_config = load_yaml(filename_hopp_config)

    # load eco inputs
    greenheart_config = load_yaml(filename_greenheart_config)

    # convert relative filepath to absolute for HOPP ingestion
    hopp_config['site']['solar_resource_file'] = convert_relative_to_absolute_path(filename_hopp_config, hopp_config['site']['solar_resource_file'])
    hopp_config['site']['wind_resource_file'] = convert_relative_to_absolute_path(filename_hopp_config, hopp_config['site']['wind_resource_file'])
    hopp_config['site']['wave_resource_file'] = convert_relative_to_absolute_path(filename_hopp_config, hopp_config['site']['wave_resource_file'])
    hopp_config['site']['grid_resource_file'] = convert_relative_to_absolute_path(filename_hopp_config, hopp_config['site']['grid_resource_file'])

    ################ load plant inputs from yaml
    if filename_orbit_config != None:
        orbit_config = orbit.load_config(filename_orbit_config)

        # print plant inputs if desired
        if verbose:
            print("\nPlant configuration:")
            for key in orbit_config.keys():
                print(key, ": ", orbit_config[key])

        # check that orbit and hopp inputs are compatible
        if (
            orbit_config["plant"]["capacity"]
            != hopp_config["technologies"]["wind"]["num_turbines"]
            * hopp_config["technologies"]["wind"]["turbine_rating_kw"]
            * 1e-3
        ):
            raise (
                ValueError("Provided ORBIT and HOPP wind plant capacities do not match")
            )

    # update floris_config file with correct input from other files
    # load floris inputs
    if (
        hopp_config["technologies"]["wind"]["model_name"] == "floris"
    ):  # TODO replace elements of the file
        if filename_floris_config is None:
            raise (ValueError("floris input file must be specified."))
        else:
            floris_config = load_yaml(filename_floris_config)
            floris_config.update({"farm": {"turbine_type": turbine_config}})
    else:
        floris_config = None

    # print turbine inputs if desired
    if verbose:
        print("\nTurbine configuration:")
        for key in turbine_config.keys():
            print(key, ": ", turbine_config[key])

    ############## provide custom layout for ORBIT and FLORIS if desired
    if filename_orbit_config != None:
        if orbit_config["plant"]["layout"] == "custom":
            # generate ORBIT config from floris layout
            for i, x in enumerate(floris_config["farm"]["layout_x"]):
                floris_config["farm"]["layout_x"][i] = x + 400

            layout_config, layout_data_location = convert_layout_from_floris_for_orbit(
                floris_config["farm"]["layout_x"],
                floris_config["farm"]["layout_y"],
                save_config=True,
            )

            # update orbit_config with custom layout
            # orbit_config = orbit.core.library.extract_library_data(orbit_config, additional_keys=layout_config)
            orbit_config["array_system_design"]["location_data"] = layout_data_location

    # if hybrid plant, adjust hybrid plant capacity to include all technologies
    total_hybrid_plant_capacity_mw = 0.0
    for tech in hopp_config["technologies"].keys():
        if tech == "grid":
            continue
        elif tech == "wind":
            total_hybrid_plant_capacity_mw += (
                hopp_config["technologies"][tech]["num_turbines"]
                * hopp_config["technologies"][tech]["turbine_rating_kw"]
                * 1e-3
            )
        elif tech == "pv":
            total_hybrid_plant_capacity_mw += (
                hopp_config["technologies"][tech]["system_capacity_kw"] * 1e-3
            )
        elif tech == "wave":
            total_hybrid_plant_capacity_mw += (
                hopp_config["technologies"][tech]["num_devices"]
                * hopp_config["technologies"][tech]["device_rating_kw"]
                * 1e-3
            )

    # initialize dict for hybrid plant
    if filename_orbit_config != None:
        if total_hybrid_plant_capacity_mw != orbit_config["plant"]["capacity"]:
            orbit_hybrid_electrical_export_config = copy.deepcopy(orbit_config)
            orbit_hybrid_electrical_export_config["plant"][
                "capacity"
            ] = total_hybrid_plant_capacity_mw
            orbit_hybrid_electrical_export_config["plant"].pop(
                "num_turbines"
            )  # allow orbit to set num_turbines later based on the new hybrid capacity and turbine rating
        else:
            orbit_hybrid_electrical_export_config = {}

    if verbose:
        print(
            f"Total hybrid plant rating calculated: {total_hybrid_plant_capacity_mw} MW"
        )

    if filename_orbit_config is None:
        orbit_config = None
        orbit_hybrid_electrical_export_config = {}

    ############## return all inputs

    return (
        hopp_config,
        greenheart_config,
        orbit_config,
        turbine_config,
        floris_config,
        orbit_hybrid_electrical_export_config,
    )


def convert_layout_from_floris_for_orbit(turbine_x, turbine_y, save_config=False):

    turbine_x_km = (np.array(turbine_x) * 1e-3).tolist()
    turbine_y_km = (np.array(turbine_y) * 1e-3).tolist()

    # initialize dict with data for turbines
    turbine_dict = {
        "id": list(range(0, len(turbine_x))),
        "substation_id": ["OSS"] * len(turbine_x),
        "name": list(range(0, len(turbine_x))),
        "longitude": turbine_x_km,
        "latitude": turbine_y_km,
        "string": [0] * len(turbine_x),  # can be left empty
        "order": [0] * len(turbine_x),  # can be left empty
        "cable_length": [0] * len(turbine_x),
        "bury_speed": [0] * len(turbine_x),
    }
    string_counter = -1
    order_counter = 0
    for i in range(0, len(turbine_x)):
        if turbine_x[i] - 400 == 0:
            string_counter += 1
            order_counter = 0

        turbine_dict["order"][i] = order_counter
        turbine_dict["string"][i] = string_counter

        order_counter += 1

    # initialize dict with substation information
    substation_dict = {
        "id": "OSS",
        "substation_id": "OSS",
        "name": "OSS",
        "longitude": np.min(turbine_x_km) - 200 * 1e-3,
        "latitude": np.average(turbine_y_km),
        "string": "",  # can be left empty
        "order": "",  # can be left empty
        "cable_length": "",
        "bury_speed": "",
    }

    # combine turbine and substation dicts
    for key in turbine_dict.keys():
        # turbine_dict[key].append(substation_dict[key])
        turbine_dict[key].insert(0, substation_dict[key])

    # add location data
    file_name = "osw_cable_layout"
    save_location = "./input/project/plant/"
    # turbine_dict["array_system_design"]["location_data"] = data_location
    if save_config:
        if not os.path.exists(save_location):
            os.makedirs(save_location)
        # create pandas data frame
        df = pd.DataFrame.from_dict(turbine_dict)

        # df.drop("index")
        df.set_index("id")

        # save to csv
        df.to_csv(save_location + file_name + ".csv", index=False)

    return turbine_dict, file_name


def visualize_plant(
    hopp_config,
    greenheart_config,
    turbine_config,
    wind_cost_outputs,
    hopp_results,
    platform_results,
    desal_results,
    h2_storage_results,
    electrolyzer_physics_results,
    design_scenario,
    colors,
    plant_design_number,
    show_plots=False,
    save_plots=False,
    output_dir="./output/",
):
    # save plant sizing to dict
    component_areas = {}

    plt.rcParams.update({"font.size": 7})

    if hopp_config["technologies"]["wind"]["model_name"] != "floris":
        raise (
            NotImplementedError(
                f"`visualize_plant()` only works with the 'floris' wind model, `model_name` \
                                  {hopp_config['technologies']['wind']['model_name']} has been specified"
            )
        )

    # set colors
    turbine_rotor_color = colors[0]
    turbine_tower_color = colors[1]
    pipe_color = colors[2]
    cable_color = colors[8]
    electrolyzer_color = colors[4]
    desal_color = colors[9]
    h2_storage_color = colors[6]
    substation_color = colors[7]
    equipment_platform_color = colors[1]
    compressor_color = colors[0]
    if hopp_config["site"]["solar"]:
        solar_color = colors[2]
    if hopp_config["site"]["wave"]:
        wave_color = colors[8]
    battery_color = colors[8]

    # set hatches
    solar_hatch = "//"
    wave_hatch = "\\\\"
    battery_hatch = "+"
    electrolyzer_hatch = "///"
    desalinator_hatch = "xxxx"

    # Views
    # offshore plant, onshore plant, offshore platform, offshore turbine

    # get plant location

    # get shore location

    # get cable/pipe locations
    if design_scenario["wind_location"] == "offshore":
        cable_array_points = (
            wind_cost_outputs.orbit_project.phases["ArraySystemDesign"].coordinates
            * 1e3
        )  # ORBIT gives coordinates in km, convert to m
        pipe_array_points = (
            wind_cost_outputs.orbit_project.phases["ArraySystemDesign"].coordinates
            * 1e3
        )  # ORBIT gives coordinates in km, convert to m

        # get turbine tower base diameter
        tower_base_diameter = wind_cost_outputs.orbit_project.config["turbine"][
            "tower"
        ]["section_diameters"][
            0
        ]  # in m
        tower_base_radius = tower_base_diameter / 2.0

        # get turbine locations
        turbine_x = (
            wind_cost_outputs.orbit_project.phases[
                "ArraySystemDesign"
            ].turbines_x.flatten()
            * 1e3
        )  # ORBIT gives coordinates in km, convert to m
        turbine_x = turbine_x[~np.isnan(turbine_x)]
        turbine_y = (
            wind_cost_outputs.orbit_project.phases[
                "ArraySystemDesign"
            ].turbines_y.flatten()
            * 1e3
        )  # ORBIT gives coordinates in km, convert to m
        turbine_y = turbine_y[~np.isnan(turbine_y)]

        # get offshore substation location and dimensions
        substation_x = (
            wind_cost_outputs.orbit_project.phases["ArraySystemDesign"].oss_x * 1e3
        )  # ORBIT gives coordinates in km, convert to m (treated as center)
        substation_y = (
            wind_cost_outputs.orbit_project.phases["ArraySystemDesign"].oss_y * 1e3
        )  # ORBIT gives coordinates in km, convert to m (treated as center)
        substation_side_length = 20  # [m] just based on a large substation (https://www.windpowerengineering.com/making-modern-offshore-substation/) since the dimensions are not available in ORBIT

        # get equipment platform location and dimensions
        equipment_platform_area = platform_results["toparea_m2"]
        equipment_platform_side_length = np.sqrt(equipment_platform_area)
        equipment_platform_x = (
            substation_x - substation_side_length - equipment_platform_side_length / 2
        )  # [m] (treated as center)
        equipment_platform_y = substation_y  # [m] (treated as center)

        # get platform equipment dimensions
        if design_scenario["electrolyzer_location"] == "turbine":
            desal_equipment_area = desal_results[
                "per_turb_equipment_footprint_m2"
            ]  # equipment_footprint_m2
        elif design_scenario["electrolyzer_location"] == "platform":
            desal_equipment_area = desal_results["equipment_footprint_m2"]
        else:
            desal_equipment_area = 0

        desal_equipment_side = np.sqrt(desal_equipment_area)

        # get pipe points
        pipe_x = np.array([substation_x - 1000, substation_x])
        pipe_y = np.array([substation_y, substation_y])

        # get cable points
        cable_x = pipe_x
        cable_y = pipe_y

    else:
        turbine_x = np.array(
            hopp_config["technologies"]["wind"]["floris_config"]["farm"]["layout_x"]
        )
        turbine_y = np.array(
            hopp_config["technologies"]["wind"]["floris_config"]["farm"]["layout_y"]
        )
        cable_array_points = []
    
    # wind farm area
    turbine_length_x = np.max(turbine_x)-np.min(turbine_x)
    turbine_length_y = np.max(turbine_y)-np.min(turbine_y)
    turbine_area = turbine_length_x * turbine_length_y

    # compressor side # not sized
    compressor_area = 25
    compressor_side = np.sqrt(compressor_area)
    ## create figure
    fig, ax = plt.subplots(2, 2, figsize=(10, 6))

    # get turbine rotor diameter
    rotor_diameter = turbine_config["rotor_diameter"]  # in m
    rotor_radius = rotor_diameter / 2.0

    # set onshore substation dimensions
    onshore_substation_x_side_length = 127.25  # [m] based on 1 acre area https://www.power-technology.com/features/making-space-for-power-how-much-land-must-renewables-use/
    onshore_substation_y_side_length = 31.8  # [m] based on 1 acre area https://www.power-technology.com/features/making-space-for-power-how-much-land-must-renewables-use/
    onshore_substation_area = onshore_substation_x_side_length * onshore_substation_y_side_length

    if greenheart_config["h2_storage"]["type"] == "pressure_vessel":
        h2_storage_area = h2_storage_results["tank_footprint_m2"]
        h2_storage_side = np.sqrt(h2_storage_area)
    else:
        h2_storage_side = 0
        h2_storage_area = 0

    electrolyzer_area = electrolyzer_physics_results["equipment_footprint_m2"]
    if design_scenario["electrolyzer_location"] == "turbine":
        electrolyzer_area /= hopp_config["technologies"]["wind"]["num_turbines"]

    electrolyzer_side = np.sqrt(electrolyzer_area)

    # set onshore origin
    onshorex = 50
    onshorey = 50

    wind_buffer = np.min(turbine_x) - (
        onshorey + 2 * rotor_diameter + electrolyzer_side
    )
    if "pv" in hopp_config["technologies"].keys():
        wind_buffer -= np.sqrt(hopp_results["hybrid_plant"].pv.footprint_area)
    if "battery" in hopp_config["technologies"].keys():
        wind_buffer -= np.sqrt(hopp_results["hybrid_plant"].battery.footprint_area)
    if wind_buffer < 50:
        onshorey += wind_buffer - 50

    if design_scenario["wind_location"] == "offshore":
        origin_x = substation_x
        origin_y = substation_y
    else:
        origin_x = 0.0
        origin_y = 0.0

    ## create figure
    if design_scenario["wind_location"] == "offshore":
        fig, ax = plt.subplots(2, 2, figsize=(10, 6))
        ax_index_plant = (0, 0)
        ax_index_detail = (1, 0)
        ax_index_wind_plant = (0, 1)
        ax_index_turbine_detail = (1, 1)
    else:
        fig, ax = plt.subplots(1, 2, figsize=(10, 6))
        ax_index_plant = 0
        ax_index_wind_plant = 0
        ax_index_detail = 1
        ax_index_turbine_detail = False

    # plot the stuff

    # onshore plant | offshore plant
    # platform/substation | turbine

    ## add turbines
    def add_turbines(ax, turbine_x, turbine_y, radius, color):
        i = 0
        for x, y in zip(turbine_x, turbine_y):
            if i == 0:
                rlabel = "Wind Turbine Rotor"
                tlabel = "Wind Turbine Tower"
                i += 1
            else:
                rlabel = None
                tlabel = None
            turbine_patch = patches.Circle(
                (x, y), radius=radius, color=color, fill=False, label=rlabel, zorder=10,
            )
            ax.add_patch(turbine_patch)

    add_turbines(
        ax[ax_index_wind_plant], turbine_x, turbine_y, rotor_radius, turbine_rotor_color
    )
    component_areas["turbine_area_m2"] = turbine_area
    # turbine_patch01_tower = patches.Circle((x, y), radius=tower_base_radius, color=turbine_tower_color, fill=False, label=tlabel, zorder=10)
    # ax[0, 1].add_patch(turbine_patch01_tower)
    if design_scenario["wind_location"] == "onshore":
        add_turbines(
            ax[ax_index_detail], turbine_x, turbine_y, rotor_radius, turbine_rotor_color
        )

    if ax_index_turbine_detail:
        # turbine_patch11_rotor = patches.Circle((turbine_x[0], turbine_y[0]), radius=rotor_radius, color=turbine_rotor_color, fill=False, label=None, zorder=10)
        tlabel = "Wind Turbine Tower"
        turbine_patch11_tower = patches.Circle(
            (turbine_x[0], turbine_y[0]),
            radius=tower_base_radius,
            color=turbine_tower_color,
            fill=False,
            label=tlabel,
            zorder=10,
        )
        # ax[1, 1].add_patch(turbine_patch11_rotor)
        ax[ax_index_turbine_detail].add_patch(turbine_patch11_tower)

    # add pipe array
    if design_scenario["transportation"] == "hvdc+pipeline" or (
        design_scenario["h2_storage_location"] != "turbine"
        and design_scenario["electrolyzer_location"] == "turbine"
    ):
        i = 0
        for point_string in pipe_array_points:
            if i == 0:
                label = "Array Pipes"
                i += 1
            else:
                label = None
            ax[0, 1].plot(
                point_string[:, 0],
                point_string[:, 1] - substation_side_length / 2,
                ":",
                color=pipe_color,
                zorder=0,
                linewidth=1,
                label=label,
            )
            ax[1, 0].plot(
                point_string[:, 0],
                point_string[:, 1] - substation_side_length / 2,
                ":",
                color=pipe_color,
                zorder=0,
                linewidth=1,
                label=label,
            )
            ax[1, 1].plot(
                point_string[:, 0],
                point_string[:, 1] - substation_side_length / 2,
                ":",
                color=pipe_color,
                zorder=0,
                linewidth=1,
                label=label,
            )

    ## add cables
    if (len(cable_array_points) > 1) and (
        design_scenario["h2_storage_location"] != "turbine"
        or design_scenario["transportation"] == "hvdc+pipeline"
    ):
        i = 0
        for point_string in cable_array_points:
            if i == 0:
                label = "Array Cables"
                i += 1
            else:
                label = None
            ax[0, 1].plot(
                point_string[:, 0],
                point_string[:, 1] + substation_side_length / 2,
                "-",
                color=cable_color,
                zorder=0,
                linewidth=1,
                label=label,
            )
            ax[1, 0].plot(
                point_string[:, 0],
                point_string[:, 1] + substation_side_length / 2,
                "-",
                color=cable_color,
                zorder=0,
                linewidth=1,
                label=label,
            )
            ax[1, 1].plot(
                point_string[:, 0],
                point_string[:, 1] + substation_side_length / 2,
                "-",
                color=cable_color,
                zorder=0,
                linewidth=1,
                label=label,
            )

    ## add offshore substation
    if design_scenario["wind_location"] == "offshore" and (
        design_scenario["h2_storage_location"] != "turbine"
        or design_scenario["transportation"] == "hvdc+pipeline"
    ):
        substation_patch01 = patches.Rectangle(
            (
                substation_x - substation_side_length,
                substation_y - substation_side_length / 2,
            ),
            substation_side_length,
            substation_side_length,
            fill=True,
            color=substation_color,
            label="Substation*",
            zorder=11,
        )
        substation_patch10 = patches.Rectangle(
            (
                substation_x - substation_side_length,
                substation_y - substation_side_length / 2,
            ),
            substation_side_length,
            substation_side_length,
            fill=True,
            color=substation_color,
            label="Substation*",
            zorder=11,
        )
        ax[0, 1].add_patch(substation_patch01)
        ax[1, 0].add_patch(substation_patch10)

        component_areas['offshore_substation_area_m2'] = substation_side_length ** 2

    ## add equipment platform
    if design_scenario["wind_location"] == "offshore" and (
        design_scenario["h2_storage_location"] == "platform"
        or design_scenario["electrolyzer_location"] == "platform"
    ):  # or design_scenario["transportation"] == "pipeline":
        equipment_platform_patch01 = patches.Rectangle(
            (
                equipment_platform_x - equipment_platform_side_length / 2,
                equipment_platform_y - equipment_platform_side_length / 2,
            ),
            equipment_platform_side_length,
            equipment_platform_side_length,
            color=equipment_platform_color,
            fill=True,
            label="Equipment Platform",
            zorder=1,
        )
        equipment_platform_patch10 = patches.Rectangle(
            (
                equipment_platform_x - equipment_platform_side_length / 2,
                equipment_platform_y - equipment_platform_side_length / 2,
            ),
            equipment_platform_side_length,
            equipment_platform_side_length,
            color=equipment_platform_color,
            fill=True,
            label="Equipment Platform",
            zorder=1,
        )
        ax[0, 1].add_patch(equipment_platform_patch01)
        ax[1, 0].add_patch(equipment_platform_patch10)

        component_areas['equipment_platform_area_m2'] = equipment_platform_area

    ## add hvdc cable
    if (
        design_scenario["transportation"] == "hvdc"
        or design_scenario["transportation"] == "hvdc+pipeline"
    ):
        ax[0, 0].plot(
            [onshorex + onshore_substation_x_side_length, 1000],
            [48, 48],
            "--",
            color=cable_color,
            label="HVDC Cable",
        )
        ax[0, 1].plot(
            [-5000, substation_x],
            [substation_y - 100, substation_y - 100],
            "--",
            color=cable_color,
            label="HVDC Cable",
            zorder=0,
        )
        ax[1, 0].plot(
            [-5000, substation_x],
            [substation_y - 2, substation_y - 2],
            "--",
            color=cable_color,
            label="HVDC Cable",
            zorder=0,
        )

    ## add onshore substation
    if (
        design_scenario["transportation"] == "hvdc"
        or design_scenario["transportation"] == "hvdc+pipeline"
    ):
        onshore_substation_patch00 = patches.Rectangle(
            (
                onshorex + 0.2 * onshore_substation_y_side_length,
                onshorey - onshore_substation_y_side_length * 1.2,
            ),
            onshore_substation_x_side_length,
            onshore_substation_y_side_length,
            fill=True,
            color=substation_color,
            label="Substation*",
            zorder=11,
        )
        ax[0, 0].add_patch(onshore_substation_patch00)

        component_areas['onshore_substation_area_m2'] = onshore_substation_area

    ## add transport pipeline
    if design_scenario["transportation"] == "colocated":
        # add hydrogen pipeline to end use
        linetype = "-."
        label = "Pipeline to Storage/End-Use"
        linewidth = 1.0

        ax[ax_index_plant].plot(
            [onshorex, -10000],
            [onshorey, onshorey],
            linetype,
            color=pipe_color,
            label=label,
            linewidth=linewidth,
            zorder=0,
        )

        ax[ax_index_detail].plot(
            [onshorex, -10000],
            [onshorey, onshorey],
            linetype,
            color=pipe_color,
            label=label,
            linewidth=linewidth,
            zorder=0,
        )
    if (
        design_scenario["transportation"] == "pipeline"
        or design_scenario["transportation"] == "hvdc+pipeline"
        or (
            design_scenario["transportation"] == "hvdc"
            and design_scenario["h2_storage_location"] == "platform"
        )
    ):
        linetype = "-."
        label = "Transport Pipeline"
        linewidth = 1.0

        ax[ax_index_plant].plot(
            [onshorex, 1000],
            [onshorey + 2, onshorey + 2],
            linetype,
            color=pipe_color,
            label=label,
            linewidth=linewidth,
            zorder=0,
        )

        if design_scenario["wind_location"] == "offshore":
            ax[ax_index_wind_plant].plot(
                [-5000, substation_x],
                [substation_y + 100, substation_y + 100],
                linetype,
                linewidth=linewidth,
                color=pipe_color,
                label=label,
                zorder=0,
            )
            ax[ax_index_detail].plot(
                [-5000, substation_x],
                [substation_y + 2, substation_y + 2],
                linetype,
                linewidth=linewidth,
                color=pipe_color,
                label=label,
                zorder=0,
            )

            if (
                design_scenario["transportation"] == "hvdc"
                or design_scenario["transportation"] == "hvdc+pipeline"
            ) and design_scenario["h2_storage_location"] == "platform":
                h2cx = onshorex - compressor_side
                h2cy = onshorey - compressor_side + 2
                h2cax = ax[ax_index_plant]
            else:
                h2cx = substation_x - substation_side_length
                h2cy = substation_y
                h2cax = ax[ax_index_detail]

        if design_scenario["wind_location"] == "onshore":
            compressor_patch01 = patches.Rectangle(
                (origin_x, origin_y),
                compressor_side,
                compressor_side,
                color=compressor_color,
                fill=None,
                label="Transport Compressor*",
                hatch="+++",
                zorder=20,
            )
            ax[ax_index_plant].add_patch(compressor_patch01)

        compressor_patch10 = patches.Rectangle(
            (h2cx, h2cy),
            compressor_side,
            compressor_side,
            color=compressor_color,
            fill=None,
            label="Transport Compressor*",
            hatch="+++",
            zorder=20,
        )
        h2cax.add_patch(compressor_patch10)

        component_areas['compressor_area_m2'] = compressor_area

    ## add plant components
    if design_scenario["electrolyzer_location"] == "onshore":
        electrolyzer_x = onshorex
        electrolyzer_y = onshorey
        electrolyzer_patch = patches.Rectangle(
            (electrolyzer_x, electrolyzer_y),
            electrolyzer_side,
            electrolyzer_side,
            color=electrolyzer_color,
            fill=None,
            label="Electrolyzer",
            zorder=20,
            hatch=electrolyzer_hatch,
        )
        ax[ax_index_plant].add_patch(electrolyzer_patch)
        component_areas['electrolyzer_area_m2'] = electrolyzer_area

        if design_scenario["wind_location"] == "onshore":
            electrolyzer_patch = patches.Rectangle(
                (onshorex - h2_storage_side, onshorey + 4),
                electrolyzer_side,
                electrolyzer_side,
                color=electrolyzer_color,
                fill=None,
                label="Electrolyzer",
                zorder=20,
                hatch=electrolyzer_hatch,
            )
            ax[ax_index_detail].add_patch(electrolyzer_patch)

    elif design_scenario["electrolyzer_location"] == "platform":
        dx = equipment_platform_x - equipment_platform_side_length / 2
        dy = equipment_platform_y - equipment_platform_side_length / 2
        e_side_y = equipment_platform_side_length
        e_side_x = electrolyzer_area / e_side_y
        d_side_y = equipment_platform_side_length
        d_side_x = desal_equipment_area / d_side_y
        electrolyzer_x = dx + d_side_x
        electrolyzer_y = dy

        electrolyzer_patch = patches.Rectangle(
            (electrolyzer_x, electrolyzer_y),
            e_side_x,
            e_side_y,
            color=electrolyzer_color,
            fill=None,
            zorder=20,
            label="Electrolyzer",
            hatch=electrolyzer_hatch,
        )
        ax[ax_index_detail].add_patch(electrolyzer_patch)
        desal_patch = patches.Rectangle(
            (dx, dy),
            d_side_x,
            d_side_y,
            color=desal_color,
            zorder=21,
            fill=None,
            label="Desalinator",
            hatch=desalinator_hatch,
        )
        ax[ax_index_detail].add_patch(desal_patch)
        component_areas['desalination_area_m2'] = desal_equipment_area

    elif design_scenario["electrolyzer_location"] == "turbine":
        electrolyzer_patch11 = patches.Rectangle(
            (turbine_x[0], turbine_y[0] + tower_base_radius),
            electrolyzer_side,
            electrolyzer_side,
            color=electrolyzer_color,
            fill=None,
            zorder=20,
            label="Electrolyzer",
            hatch=electrolyzer_hatch,
        )
        ax[ax_index_turbine_detail].add_patch(electrolyzer_patch11)
        desal_patch11 = patches.Rectangle(
            (turbine_x[0] - desal_equipment_side, turbine_y[0] + tower_base_radius),
            desal_equipment_side,
            desal_equipment_side,
            color=desal_color,
            zorder=21,
            fill=None,
            label="Desalinator",
            hatch=desalinator_hatch,
        )
        ax[ax_index_turbine_detail].add_patch(desal_patch11)
        component_areas['desalination_area_m2'] = desal_equipment_area
        i = 0
        for x, y in zip(turbine_x, turbine_y):
            if i == 0:
                elable = "Electrolyzer"
                dlabel = "Desalinator"
            else:
                elable = None
                dlabel = None
            electrolyzer_patch01 = patches.Rectangle(
                (x, y + tower_base_radius),
                electrolyzer_side,
                electrolyzer_side,
                color=electrolyzer_color,
                fill=None,
                zorder=20,
                label=elable,
                hatch=electrolyzer_hatch,
            )
            desal_patch01 = patches.Rectangle(
                (x - desal_equipment_side, y + tower_base_radius),
                desal_equipment_side,
                desal_equipment_side,
                color=desal_color,
                zorder=21,
                fill=None,
                label=dlabel,
                hatch=desalinator_hatch,
            )
            ax[ax_index_wind_plant].add_patch(electrolyzer_patch01)
            ax[ax_index_wind_plant].add_patch(desal_patch01)
            i += 1

    h2_storage_hatch = "\\\\\\"
    if design_scenario["h2_storage_location"] == "onshore" and (
        greenheart_config["h2_storage"]["type"] != "none"
    ):
        h2_storage_patch = patches.Rectangle(
            (onshorex - h2_storage_side, onshorey - h2_storage_side - 2),
            h2_storage_side,
            h2_storage_side,
            color=h2_storage_color,
            fill=None,
            label="H$_2$ Storage",
            hatch=h2_storage_hatch,
        )
        ax[ax_index_plant].add_patch(h2_storage_patch)
        component_areas["h2_storage_area_m2"] = h2_storage_area

        if design_scenario["wind_location"] == "onshore":
            h2_storage_patch = patches.Rectangle(
                (onshorex - h2_storage_side, onshorey - h2_storage_side - 2),
                h2_storage_side,
                h2_storage_side,
                color=h2_storage_color,
                fill=None,
                label="H$_2$ Storage",
                hatch=h2_storage_hatch,
            )
            ax[ax_index_detail].add_patch(h2_storage_patch)
            component_areas["h2_storage_area_m2"] = h2_storage_area
    elif design_scenario["h2_storage_location"] == "platform" and (
        greenheart_config["h2_storage"]["type"] != "none"
    ):
        s_side_y = equipment_platform_side_length
        s_side_x = h2_storage_area / s_side_y
        sx = equipment_platform_x - equipment_platform_side_length / 2
        sy = equipment_platform_y - equipment_platform_side_length / 2
        if design_scenario["electrolyzer_location"] == "platform":
            sx += equipment_platform_side_length - s_side_x

        h2_storage_patch = patches.Rectangle(
            (sx, sy),
            s_side_x,
            s_side_y,
            color=h2_storage_color,
            fill=None,
            label="H$_2$ Storage",
            hatch=h2_storage_hatch,
        )
        ax[ax_index_detail].add_patch(h2_storage_patch)
        component_areas["h2_storage_area_m2"] = h2_storage_area

    elif design_scenario["h2_storage_location"] == "turbine":

        if greenheart_config["h2_storage"]["type"] == "turbine":
            h2_storage_patch = patches.Circle(
                (turbine_x[0], turbine_y[0]),
                radius=tower_base_diameter / 2,
                color=h2_storage_color,
                fill=None,
                label="H$_2$ Storage",
                hatch=h2_storage_hatch,
            )
            ax[ax_index_turbine_detail].add_patch(h2_storage_patch)
            component_areas["h2_storage_area_m2"] = h2_storage_area
            i = 0
            for x, y in zip(turbine_x, turbine_y):
                if i == 0:
                    slable = "H$_2$ Storage"
                else:
                    slable = None
                h2_storage_patch = patches.Circle(
                    (x, y),
                    radius=tower_base_diameter / 2,
                    color=h2_storage_color,
                    fill=None,
                    label=None,
                    hatch=h2_storage_hatch,
                )
                ax[ax_index_wind_plant].add_patch(h2_storage_patch)
        elif greenheart_config["h2_storage"]["type"] == "pressure_vessel":
            h2_storage_side = np.sqrt(
                h2_storage_area / greenheart_config["plant"]["num_turbines"]
            )
            h2_storage_patch = patches.Rectangle(
                (
                    turbine_x[0] - h2_storage_side - desal_equipment_side,
                    turbine_y[0] + tower_base_radius,
                ),
                width=h2_storage_side,
                height=h2_storage_side,
                color=h2_storage_color,
                fill=None,
                label="H$_2$ Storage",
                hatch=h2_storage_hatch,
            )
            ax[ax_index_turbine_detail].add_patch(h2_storage_patch)
            component_areas["h2_storage_area_m2"] = h2_storage_area
            i = 0
            for x, y in zip(turbine_x, turbine_y):
                if i == 0:
                    slable = "H$_2$ Storage"
                else:
                    slable = None
                h2_storage_patch = patches.Rectangle(
                    (
                        turbine_x[i] - h2_storage_side - desal_equipment_side,
                        turbine_y[i] + tower_base_radius,
                    ),
                    width=h2_storage_side,
                    height=h2_storage_side,
                    color=h2_storage_color,
                    fill=None,
                    label=slable,
                    hatch=h2_storage_hatch,
                )
                ax[ax_index_wind_plant].add_patch(h2_storage_patch)
                i += 1

    ## add battery
    if "battery" in hopp_config["technologies"].keys():
        component_areas['battery_area_m2'] = hopp_results["hybrid_plant"].battery.footprint_area
        if design_scenario["battery_location"] == "onshore":
            battery_side_y = np.sqrt(
                hopp_results["hybrid_plant"].battery.footprint_area
            )
            battery_side_x = battery_side_y

            batteryx = electrolyzer_x

            batteryy = electrolyzer_y + electrolyzer_side + 10

            battery_patch = patches.Rectangle(
                (batteryx, batteryy),
                battery_side_x,
                battery_side_y,
                color=battery_color,
                fill=None,
                label="Battery Array",
                hatch=battery_hatch,
            )
            ax[ax_index_plant].add_patch(battery_patch)

            if design_scenario["wind_location"] == "onshore":

                battery_patch = patches.Rectangle(
                    (batteryx, batteryy),
                    battery_side_x,
                    battery_side_y,
                    color=battery_color,
                    fill=None,
                    label="Battery Array",
                    hatch=battery_hatch,
                )
                ax[ax_index_detail].add_patch(battery_patch)

        elif design_scenario["battery_location"] == "platform":
            battery_side_y = equipment_platform_side_length
            battery_side_x = (
                hopp_results["hybrid_plant"].battery.footprint_area / battery_side_y
            )

            batteryx = equipment_platform_x - equipment_platform_side_length / 2
            batteryy = equipment_platform_y - equipment_platform_side_length / 2

            battery_patch = patches.Rectangle(
                (batteryx, batteryy),
                battery_side_x,
                battery_side_y,
                color=battery_color,
                fill=None,
                label="Battery Array",
                hatch=battery_hatch,
            )
            ax[ax_index_detail].add_patch(battery_patch)   

    else:
        battery_side_y = 0.0
        battery_side_x = 0.0   
    
    ## add solar
    if hopp_config["site"]["solar"]:
        component_areas['pv_area_m2'] = hopp_results["hybrid_plant"].pv.footprint_area
        if design_scenario["pv_location"] == "offshore":
            solar_side_y = equipment_platform_side_length
            solar_side_x = hopp_results["hybrid_plant"].pv.footprint_area / solar_side_y

            solarx = equipment_platform_x - equipment_platform_side_length / 2
            solary = equipment_platform_y - equipment_platform_side_length / 2

            solar_patch = patches.Rectangle(
                (solarx, solary),
                solar_side_x,
                solar_side_y,
                color=solar_color,
                fill=None,
                label="Solar Array",
                hatch=solar_hatch,
            )
            ax[ax_index_detail].add_patch(solar_patch)
        else:
            solar_side_y = np.sqrt(hopp_results["hybrid_plant"].pv.footprint_area)
            solar_side_x = hopp_results["hybrid_plant"].pv.footprint_area / solar_side_y

            solarx = electrolyzer_x

            solary = electrolyzer_y + electrolyzer_side + 10

            if "battery" in hopp_config["technologies"].keys():
                solary += battery_side_y + 10

            solar_patch = patches.Rectangle(
                (solarx, solary),
                solar_side_x,
                solar_side_y,
                color=solar_color,
                fill=None,
                label="Solar Array",
                hatch=solar_hatch,
            )

            ax[ax_index_plant].add_patch(solar_patch)

            solar_patch = patches.Rectangle(
                (solarx, solary),
                solar_side_x,
                solar_side_y,
                color=solar_color,
                fill=None,
                label="Solar Array",
                hatch=solar_hatch,
            )

            ax[ax_index_detail].add_patch(solar_patch)
    else:
        solar_side_x = 0.0
        solar_side_y = 0.0

    ## add wave
    if hopp_config["site"]["wave"]:
        # get wave generation area geometry
        num_devices = hopp_config["technologies"]["wave"]["num_devices"]
        distance_to_shore = (
            hopp_config["technologies"]["wave"]["cost_inputs"]["distance_to_shore"]
            * 1e3
        )
        number_rows = hopp_config["technologies"]["wave"]["cost_inputs"]["number_rows"]
        device_spacing = hopp_config["technologies"]["wave"]["cost_inputs"][
            "device_spacing"
        ]
        row_spacing = hopp_config["technologies"]["wave"]["cost_inputs"]["row_spacing"]

        # calculate wave generation area dimenstions
        wave_side_y = device_spacing * np.ceil(num_devices / number_rows)
        wave_side_x = row_spacing * (number_rows)
        wave_area = wave_side_x * wave_side_y
        component_areas['wave_area_m2'] = wave_area

        # generate wave generation patch
        wavex = substation_x - wave_side_x
        wavey = substation_y + distance_to_shore
        wave_patch = patches.Rectangle(
            (wavex, wavey),
            wave_side_x,
            wave_side_y,
            color=wave_color,
            fill=None,
            label="Wave Array",
            hatch=wave_hatch,
            zorder=1,
        )
        ax[ax_index_wind_plant].add_patch(wave_patch)

        # add electrical transmission for wave
        wave_export_cable_coords_x = [substation_x, substation_x]
        wave_export_cable_coords_y = [substation_y, substation_y + distance_to_shore]

        ax[ax_index_wind_plant].plot(
            wave_export_cable_coords_x,
            wave_export_cable_coords_y,
            cable_color,
            zorder=0,
        )
        ax[ax_index_detail].plot(
            wave_export_cable_coords_x,
            wave_export_cable_coords_y,
            cable_color,
            zorder=0,
        )

    if design_scenario["wind_location"] == "offshore":
        allpoints = cable_array_points.flatten()
    else:
        allpoints = turbine_x

    allpoints = allpoints[~np.isnan(allpoints)]

    if design_scenario["wind_location"] == "offshore":
        roundto = -2
        ax[ax_index_plant].set(
            xlim=[
                round(np.min(onshorex - 100), ndigits=roundto),
                round(
                    np.max(
                        onshorex
                        + onshore_substation_x_side_length
                        + electrolyzer_side
                        + 200
                    ),
                    ndigits=roundto,
                ),
            ],
            ylim=[
                round(np.min(onshorey - 100), ndigits=roundto),
                round(
                    np.max(
                        onshorey
                        + battery_side_y
                        + electrolyzer_side
                        + solar_side_y
                        + 100
                    ),
                    ndigits=roundto,
                ),
            ],
        )
        ax[ax_index_plant].set(aspect="equal")
    else:
        roundto = -3
        ax[ax_index_plant].set(
            xlim=[
                round(np.min(allpoints - 6000), ndigits=roundto),
                round(np.max(allpoints + 6000), ndigits=roundto),
            ],
            ylim=[
                round(np.min(onshorey - 1000), ndigits=roundto),
                round(np.max(turbine_y + 4000), ndigits=roundto),
            ],
        )
        ax[ax_index_plant].autoscale()
        ax[ax_index_plant].set(aspect="equal")
        ax[ax_index_plant].xaxis.set_major_locator(ticker.MultipleLocator(2000))
        ax[ax_index_plant].yaxis.set_major_locator(ticker.MultipleLocator(1000))

    roundto = -3
    ax[ax_index_wind_plant].set(
        xlim=[
            round(np.min(allpoints - 6000), ndigits=roundto),
            round(np.max(allpoints + 6000), ndigits=roundto),
        ],
        ylim=[
            round((np.min([np.min(turbine_y), onshorey]) - 1000), ndigits=roundto),
            round(np.max(turbine_y + 4000), ndigits=roundto),
        ],
    )
    # ax[ax_index_wind_plant].autoscale()
    ax[ax_index_wind_plant].set(aspect="equal")
    ax[ax_index_wind_plant].xaxis.set_major_locator(ticker.MultipleLocator(5000))
    ax[ax_index_wind_plant].yaxis.set_major_locator(ticker.MultipleLocator(1000))

    if design_scenario["wind_location"] == "offshore":
        roundto = -2
        ax[ax_index_detail].set(
            xlim=[
                round(origin_x - 400, ndigits=roundto),
                round(origin_x + 100, ndigits=roundto),
            ],
            ylim=[
                round(origin_y - 200, ndigits=roundto),
                round(origin_y + 200, ndigits=roundto),
            ],
        )
        ax[ax_index_detail].set(aspect="equal")
    else:
        roundto = -2

        if "pv" in hopp_config["technologies"].keys():
            xmax = round(
                np.max([onshorex + 510, solarx + solar_side_x + 100]), ndigits=roundto
            )
            ymax = round(solary + solar_side_y + 100, ndigits=roundto)
        else:
            xmax = round(np.max([onshorex + 510, 100]), ndigits=roundto)
            ymax = round(100, ndigits=roundto)
        ax[ax_index_detail].set(
            xlim=[round(onshorex - 10, ndigits=roundto), xmax,],
            ylim=[round(onshorey - 200, ndigits=roundto), ymax,],
        )
        ax[ax_index_detail].set(aspect="equal")

    if design_scenario["wind_location"] == "offshore":
        tower_buffer0 = 10
        tower_buffer1 = 10
        roundto = -1
        ax[ax_index_turbine_detail].set(
            xlim=[
                round(
                    turbine_x[0] - tower_base_radius - tower_buffer0 - 50,
                    ndigits=roundto,
                ),
                round(
                    turbine_x[0] + tower_base_radius + 3 * tower_buffer1,
                    ndigits=roundto,
                ),
            ],
            ylim=[
                round(
                    turbine_y[0] - tower_base_radius - 2 * tower_buffer0,
                    ndigits=roundto,
                ),
                round(
                    turbine_y[0] + tower_base_radius + 4 * tower_buffer1,
                    ndigits=roundto,
                ),
            ],
        )
        ax[ax_index_turbine_detail].set(aspect="equal")
        ax[ax_index_turbine_detail].xaxis.set_major_locator(ticker.MultipleLocator(10))
        ax[ax_index_turbine_detail].yaxis.set_major_locator(ticker.MultipleLocator(10))
        # ax[0,1].legend(frameon=False)
        # ax[0,1].axis('off')

    if design_scenario["wind_location"] == "offshore":
        labels = [
            "(a) Onshore plant",
            "(b) Offshore plant",
            "(c) Equipment platform and substation",
            "(d) NW-most wind turbine",
        ]
    else:
        labels = ["(a) Full plant", "(b) Non-wind plant detail"]
    for axi, label in zip(ax.flatten(), labels):
        axi.legend(frameon=False, ncol=2)  # , ncol=2, loc="best")
        axi.set(xlabel="Easting (m)", ylabel="Northing (m)")
        axi.set_title(label, loc="left")
        # axi.spines[['right', 'top']].set_visible(False)

    ## save the plot
    plt.tight_layout()
    savepaths = [
            output_dir + "figures/layout/",
            output_dir + "data/",
        ]
    if save_plots:
        for savepath in savepaths:
            if not os.path.exists(savepath):
                os.makedirs(savepath)
        plt.savefig(
            savepaths[0] + "plant_layout_%i.png" % (plant_design_number), transparent=True
        )
        
        df = pd.DataFrame([component_areas])
        df.to_csv(savepaths[1] + "component_areas_layout_%i.csv" % (plant_design_number), index=False)

    if show_plots:
        plt.show()
    return 0


def save_energy_flows(
    hybrid_plant: HoppInterface.system, 
    electrolyzer_physics_results, 
    solver_results, 
    hours, 
    h2_storage_results,
    ax=None, 
    simulation_length=8760, 
    output_dir="./output/",
):

    

    if ax == None:
        fig, ax = plt.subplots(1)

    output = {}
    if hybrid_plant.pv:
        solar_plant_power = np.array(
            hybrid_plant.pv.generation_profile[0:simulation_length]
        )
        output.update({"pv generation [kW]": solar_plant_power})
    if hybrid_plant.wind:
        wind_plant_power = np.array(
            hybrid_plant.wind.generation_profile[0:simulation_length]
        )
        output.update({"wind generation [kW]": wind_plant_power})
    if hybrid_plant.wave:
        wave_plant_power = np.array(
            hybrid_plant.wave.generation_profile[0:simulation_length]
        )
        output.update({"wave generation [kW]": wave_plant_power})
    if hybrid_plant.battery:
        battery_power_out_mw = hybrid_plant.battery.outputs.P 
        output.update({"battery discharge [kW]": [(int(p>0))*p*1E3 for p in battery_power_out_mw]}) # convert from MW to kW and extract only discharging
        output.update({"battery charge [kW]": [-(int(p<0))*p*1E3 for p in battery_power_out_mw]}) # convert from MW to kW and extract only charging
        output.update({"battery state of charge [%]": hybrid_plant.battery.outputs.dispatch_SOC})

    output.update({"total accessory power required [kW]": solver_results[0]})
    output.update({"grid energy usage hourly [kW]": [solver_results[1]]*simulation_length})
    output.update({"desal energy hourly [kW]": [solver_results[2]]*simulation_length})
    output.update({"electrolyzer energy hourly [kW]": electrolyzer_physics_results["power_to_electrolyzer_kw"]})
    output.update({"electrolyzer bop energy hourly [kW]":solver_results[5]})
    output.update({"transport compressor energy hourly [kW]": [solver_results[3]]*simulation_length})
    output.update({"storage energy hourly [kW]": [solver_results[4]]*simulation_length})
    output.update({"h2 production hourly [kg]": electrolyzer_physics_results["H2_Results"]["Hydrogen Hourly Production [kg/hr]"]})
    if "hydrogen_storage_soc" in h2_storage_results:
        output.update({"hydrogen storage SOC [kg]": h2_storage_results["hydrogen_storage_soc"]})
    
    df = pd.DataFrame.from_dict(output)

    filepath = os.path.abspath(output_dir + "data/production/")

    if not os.path.exists(filepath):
        os.makedirs(filepath)

    df.to_csv(os.path.join(filepath, "energy_flows.csv"))

    return output


# set up function to post-process HOPP results
def post_process_simulation(
    lcoe,
    lcoh,
    pf_lcoh,
    pf_lcoe,
    hopp_results,
    electrolyzer_physics_results,
    hopp_config,
    greenheart_config,
    orbit_config,
    turbine_config,
    h2_storage_results,
    capex_breakdown,
    opex_breakdown,
    wind_cost_results,
    platform_results,
    desal_results,
    design_scenario,
    plant_design_number,
    incentive_option,
    solver_results=[],
    show_plots=False,
    save_plots=False,
    verbose=False,
    output_dir="./output/",
):  # , lcoe, lcoh, lcoh_with_grid, lcoh_grid_only):
    # colors (official NREL color palette https://brand.nrel.gov/content/index/guid/color_palette?parent=61)
    colors = [
        "#0079C2",
        "#00A4E4",
        "#F7A11A",
        "#FFC423",
        "#5D9732",
        "#8CC63F",
        "#5E6A71",
        "#D1D5D8",
        "#933C06",
        "#D9531E",
    ]
    # load saved results

    # post process results
    if verbose:
        print("LCOE: ", round(lcoe * 1e3, 2), "$/MWh")
        print("LCOH: ", round(lcoh, 2), "$/kg")
        print(
            "hybrid electricity plant capacity factor: ",
            round(
                np.sum(hopp_results["combined_hybrid_power_production_hopp"])
                / (hopp_results["hybrid_plant"].system_capacity_kw.hybrid * 365 * 24),
                2,
            ),
        )
        print(
            "electrolyzer capacity factor: ",
            round(
                np.sum(electrolyzer_physics_results["power_to_electrolyzer_kw"])
                * 1e-3
                / (greenheart_config["electrolyzer"]["rating"] * 365 * 24),
                2,
            ),
        )
        print(
            "Electorlyzer CAPEX installed $/kW: ",
            round(
                capex_breakdown["electrolyzer"]
                / (greenheart_config["electrolyzer"]["rating"] * 1e3),
                2,
            ),
        )

    if show_plots or save_plots:
        visualize_plant(
            hopp_config,
            greenheart_config,
            turbine_config,
            wind_cost_results,
            hopp_results,
            platform_results,
            desal_results,
            h2_storage_results,
            electrolyzer_physics_results,
            design_scenario,
            colors,
            plant_design_number,
            show_plots=show_plots,
            save_plots=save_plots,
            output_dir=output_dir,
        )
    savepaths = [
        output_dir + "data/",
        output_dir + "data/lcoe/",
        output_dir + "data/lcoh/",
    ]
    for sp in savepaths:
        if not os.path.exists(sp):
            os.makedirs(sp)

    pf_lcoh.get_cost_breakdown().to_csv(
        savepaths[2]
        + "cost_breakdown_lcoh_design%i_incentive%i_%sstorage.csv"
        % (
            plant_design_number,
            incentive_option,
            greenheart_config["h2_storage"]["type"],
        )
    )
    pf_lcoe.get_cost_breakdown().to_csv(
        savepaths[1]
        + "cost_breakdown_lcoe_design%i_incentive%i_%sstorage.csv"
        % (
            plant_design_number,
            incentive_option,
            greenheart_config["h2_storage"]["type"],
        )
    )

    # create dataframe for saving all the stuff
    greenheart_config["design_scenario"] = design_scenario
    greenheart_config["plant_design_number"] = plant_design_number
    greenheart_config["incentive_options"] = incentive_option

    # save power usage data
    if len(solver_results) > 0:
        hours = len(hopp_results["combined_hybrid_power_production_hopp"])
        annual_energy_breakdown = {
            "electricity_generation_kwh": sum(
                hopp_results["combined_hybrid_power_production_hopp"]
            ),
            "electrolyzer_kwh": sum(
                electrolyzer_physics_results["power_to_electrolyzer_kw"]
            ),
            "renewable_kwh": sum(solver_results[0]),
            "grid_power_kwh": solver_results[1] * hours,
            "desal_kwh": solver_results[2] * hours,
            "h2_transport_compressor_power_kwh": solver_results[3] * hours,
            "h2_storage_power_kwh": solver_results[4] * hours,
            "electrolyzer_bop_energy_kwh": sum(solver_results[5])
        }


    ######################### save detailed ORBIT cost information
    if wind_cost_results.orbit_project:
        _, orbit_capex_breakdown, wind_capex_multiplier = adjust_orbit_costs(
            orbit_project=wind_cost_results.orbit_project,
            greenheart_config=greenheart_config,
        )

        # orbit_capex_breakdown["Onshore Substation"] = orbit_project.phases["ElectricalDesign"].onshore_cost
        # discount ORBIT cost information
        for key in orbit_capex_breakdown:
            orbit_capex_breakdown[key] = -npf.fv(
                greenheart_config["finance_parameters"]["costing_general_inflation"],
                greenheart_config["project_parameters"]["cost_year"]
                - greenheart_config["finance_parameters"]["discount_years"]["wind"],
                0.0,
                orbit_capex_breakdown[key],
            )

        # save ORBIT cost information
        ob_df = pd.DataFrame(orbit_capex_breakdown, index=[0]).transpose()
        savedir = output_dir + "data/orbit_costs/"
        if not os.path.exists(savedir):
            os.makedirs(savedir)
        ob_df.to_csv(
            savedir
            + "orbit_cost_breakdown_lcoh_design%i_incentive%i_%sstorage.csv"
            % (
                plant_design_number,
                incentive_option,
                greenheart_config["h2_storage"]["type"],
            )
        )
        ###############################

        ###################### Save export system breakdown from ORBIT ###################

        _, orbit_capex_breakdown, wind_capex_multiplier = adjust_orbit_costs(
            orbit_project=wind_cost_results.orbit_project,
            greenheart_config=greenheart_config,
        )

        onshore_substation_costs = (
            wind_cost_results.orbit_project.phases["ElectricalDesign"].onshore_cost
            * wind_capex_multiplier
        )

        orbit_capex_breakdown["Export System Installation"] -= onshore_substation_costs

        orbit_capex_breakdown[
            "Onshore Substation and Installation"
        ] = onshore_substation_costs

        # discount ORBIT cost information
        for key in orbit_capex_breakdown:
            orbit_capex_breakdown[key] = -npf.fv(
                greenheart_config["finance_parameters"]["costing_general_inflation"],
                greenheart_config["project_parameters"]["cost_year"]
                - greenheart_config["finance_parameters"]["discount_years"]["wind"],
                0.0,
                orbit_capex_breakdown[key],
            )

        # save ORBIT cost information
        ob_df = pd.DataFrame(orbit_capex_breakdown, index=[0]).transpose()
        savedir = output_dir + "data/orbit_costs/"
        if not os.path.exists(savedir):
            os.makedirs(savedir)
        ob_df.to_csv(
            savedir
            + "orbit_cost_breakdown_with_onshore_substation_lcoh_design%i_incentive%i_%sstorage.csv"
            % (
                plant_design_number,
                incentive_option,
                greenheart_config["h2_storage"]["type"],
            )
        )

    ##################################################################################
    if (
        hasattr(hopp_results["hybrid_plant"], "dispatch_builder")
        and hopp_results["hybrid_plant"].battery
    ):
        savedir = output_dir + "figures/production/"
        if not os.path.exists(savedir):
            os.makedirs(savedir)
        plot_tools.plot_generation_profile(
            hopp_results["hybrid_plant"],
            start_day=0,
            n_days=10,
            plot_filename=os.path.abspath(savedir + "generation_profile.pdf"),
            font_size=14,
            power_scale=1 / 1000,
            solar_color="r",
            wind_color="b",
            # wave_color="g",
            discharge_color="b",
            charge_color="r",
            gen_color="g",
            price_color="r",
            # show_price=False,
        )
    else:
        print(
            "generation profile not plotted because HoppInterface does not have a "
            "'dispatch_builder'"
        )

    # save production information
    hourly_energy_breakdown = save_energy_flows(
        hopp_results["hybrid_plant"],
        electrolyzer_physics_results,
        solver_results,
        hours,
        h2_storage_results,
        output_dir=output_dir
    )

    # save hydrogen information
    key = "Hydrogen Hourly Production [kg/hr]"
    np.savetxt(
        output_dir + "h2_usage",
        electrolyzer_physics_results["H2_Results"][key],
        header="# " + key
    )

    return annual_energy_breakdown, hourly_energy_breakdown



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 18 09:06:19 2022

@author: ereznic2
"""
import os
import numpy as np
import pandas as pd

import hopp.to_organize.hydrogen_steel_pipe_cost_functions as hydrogen_steel_pipe_cost_functions

#parent_path = os.path.abspath('')

def hydrogen_steel_pipeline_cost_analysis(parent_path,turbine_model,hydrogen_max_hourly_production_kg,site_name):
    pipe_info_dir = parent_path + '/H2_Analysis/'

    pipeline_info = pd.read_csv(pipe_info_dir+'/Pipeline_info.csv',header = 0,sep=',')

    pipeline_info = pipeline_info.loc[pipeline_info['Site']==site_name].reset_index().drop(['index'],axis=1)

    total_plant_hydrogen_production_kgpsec = hydrogen_max_hourly_production_kg/3600

    pipe_diameter_min_mm = []
    for j in range(pipeline_info.shape[0]):
        #j=0
        flowrate_per_turbine = total_plant_hydrogen_production_kgpsec/pipeline_info.loc[j,'Total number of turbines']
        pipe_mass_flow_kgpsec = flowrate_per_turbine*pipeline_info.loc[j,'Number of turbines in each pipe']
        # Decide pipe inlet pressure based on where the pipe starts
        if pipeline_info.loc[j,'Pipe start point'] == 'Turbine':
            pressure_inlet_bar = 40
        elif pipeline_info.loc[j,'Pipe start point'] == 'Pipe':
            pressure_inlet_bar = 20

        if pipeline_info.loc[j,'Pipe end point'] == 'Central':
            pressure_outlet_bar =5
        elif pipeline_info.loc[j,'Pipe end point'] == 'Pipe':
            pressure_outlet_bar = 20

        pipe_diameter_min_mm.append(25.4*hydrogen_steel_pipe_cost_functions.get_diameter_of_pipe(pipeline_info.loc[j,'Length of Pipe in Arm (m)'],pipe_mass_flow_kgpsec,pressure_inlet_bar,pressure_outlet_bar))

    # Design specifications for the pipes
    grade = 'X52'
    design_option = 'B'
    location_class = 1
    joint_factor = 1
    design_pressure_asme = pressure_inlet_bar/10

    # Calculate minimum DN, schedule, and associated thickness
    pipe_min_DN = []
    pipe_outer_diameter = []
    pipe_inner_diameter = []
    pipe_min_schedule = []
    pipe_min_thickness = []

    # Loop through pipes in the network
    for min_inner_diameter in pipe_diameter_min_mm:

        dn,outer_diameter,schedule_minviable,thickness_of_schedule=hydrogen_steel_pipe_cost_functions.get_dn_schedule_of_pipe(pipe_info_dir,grade,design_option,location_class,joint_factor,min_inner_diameter,design_pressure_asme)

        inner_diameter = outer_diameter - 2*thickness_of_schedule

        pipe_min_DN.append(dn)
        pipe_outer_diameter.append(outer_diameter)
        pipe_inner_diameter.append(inner_diameter)
        pipe_min_schedule.append(schedule_minviable)
        pipe_min_thickness.append(thickness_of_schedule)

    # Economic analysis
    steel_costs_per_kg_all_grades = pd.read_csv(pipe_info_dir+'steel_costs_per_kg.csv',index_col = None,header = 0)
    steel_cost_per_kg = steel_costs_per_kg_all_grades.loc[steel_costs_per_kg_all_grades['Grade']==grade,'Price [$/kg]'].to_list()[0]
    density_steel = 7840

    cpi_2004 = 251.1
    cpi_2019 = 255.7

    cpi_ratio = cpi_2019/cpi_2004

    pipe_material_cost_bymass_USD = []
    pipe_material_cost_USD = []
    pipe_misc_cost_USD = []
    pipe_labor_cost_USD = []
    pipe_row_cost_USD = []
    pipe_total_cost_USD = []

    #   ANL costing coefficients for all regions. Material cost is ignored as internal methods are used
    anl_coefs_regional = {'GP':{'labor':[10406,0.20953,-0.08419],'misc':[4944,0.17351,-0.07621],'ROW':[2751,-0.28294,0.00731],'Material':[5813,0.31599,-0.00376]},
                            'NE':{'labor':[249131,-0.33162,-0.17892],'misc':[65990,-0.29673,-0.06856],'ROW':[83124,-0.66357,-0.07544],'Material':[10409,0.296847,-0.07257]},
                            'MA':{'labor':[43692,0.05683,-0.10108],'misc':[14616,0.16354,-0.16186],'ROW':[1942,0.17394,-0.01555],'Material':[9113,0.279875,-0.00840]},
                            'GL':{'labor':[58154,-0.14821,-0.10596],'misc':[41238,-0.34751,-0.11104],'ROW':[14259,-0.65318,0.06865],'Material':[8971,0.255012,-0.03138]},
                            'RM':{'labor':[10406,0.20953,-0.08419],'misc':[4944,0.17351,-0.07621],'ROW':[2751,-0.28294,0.00731],'Material':[5813,0.31599,-0.00376]},
                            'SE':{'labor':[32094,0.06110,-0.14828],'misc':[11270,0.19077,-0.13669],'ROW':[9531,-0.37284,0.02616],'Material':[6207,0.38224,-0.05211]},
                            'PN':{'labor':[32094,0.06110,-0.14828],'misc':[11270,0.19077,-0.13669],'ROW':[9531,-0.37284,0.02616],'Material':[6207,0.38224,-0.05211]},
                            'SW':{'labor':[95295,-0.53848,0.03070],'misc':[19211,-0.14178,-0.04697],'ROW':[72634,-1.07566,0.05284],'Material':[5605,0.41642,-0.06441]},
                            'CA':{'labor':[95295,-0.53848,0.03070],'misc':[19211,-0.14178,-0.04697],'ROW':[72634,-1.07566,0.05284],'Material':[5605,0.41642,-0.06441]}}

    region={'IN':'GL','TX':'SW','IA':'GP','MS':'SE','MN':'GP','WY':'RM'}
    anl_coefs = anl_coefs_regional[region[site_name]]
    small_positive = 1e-7 # This allows solution and file output writing when length of a given DN is set to zero, but usually this is a sign of an issue somewhere
    for j in range(len(pipe_min_DN)):
       # j = 0

        # Calculate pipe material cost based on pipe volume
        pipe_volume = np.pi/4*((0.001*pipe_outer_diameter[j])**2 - (0.001*pipe_inner_diameter[j])**2)*pipeline_info.loc[j,'Length of Pipe in Arm (m)']
        pipe_mass_kg = density_steel*pipe_volume
        pipe_material_cost_bymass_USD.append(steel_cost_per_kg*pipe_mass_kg*pipeline_info.loc[j,'Number of such pipes needed'])

        # Calculate costs using Parker 2004 data adjusted to 2019 (two methods)
        pipeline_length_miles = pipeline_info.loc[j,'Number of such pipes needed']*pipeline_info.loc[j,'Length of Pipe in Arm (m)']*0.621371/1000
        pipe_diam_in = pipe_outer_diameter[j]/25.4

        pipe_material_cost_USD.append(cpi_ratio*anl_coefs['Material'][0]*(pipe_diam_in**anl_coefs['Material'][1])*(pipeline_length_miles**anl_coefs['Material'][2]+small_positive)*pipe_diam_in*pipeline_length_miles)
        pipe_labor_cost_USD.append(cpi_ratio*anl_coefs['labor'][0]*(pipe_diam_in**anl_coefs['labor'][1])*(pipeline_length_miles**anl_coefs['labor'][2]+small_positive)*pipe_diam_in*pipeline_length_miles)
        pipe_misc_cost_USD.append(cpi_ratio*anl_coefs['misc'][0]*(pipe_diam_in**anl_coefs['misc'][1])*(pipeline_length_miles**anl_coefs['misc'][2]+small_positive)*pipe_diam_in*pipeline_length_miles)
        pipe_row_cost_USD.append( cpi_ratio*anl_coefs['ROW'][0]*(pipe_diam_in**anl_coefs['ROW'][1])*(pipeline_length_miles**anl_coefs['ROW'][2]+small_positive)*pipe_diam_in*pipeline_length_miles)

        pipe_total_cost_USD.append(pipe_material_cost_USD[j] + pipe_misc_cost_USD[j] + pipe_labor_cost_USD[j] + pipe_row_cost_USD[j])

    pipe_material_cost_USD = pd.DataFrame(pipe_material_cost_USD,columns = ['Total material cost ($)'])
    pipe_misc_cost_USD = pd.DataFrame(pipe_misc_cost_USD,columns = ['Total misc cost ($)'])
    pipe_labor_cost_USD = pd.DataFrame(pipe_labor_cost_USD,columns = ['Total labor cost ($)'])
    pipe_row_cost_USD = pd.DataFrame(pipe_row_cost_USD,columns = ['Total ROW cost ($)'])

    pipe_material_cost_bymass_USD = pd.DataFrame(pipe_material_cost_bymass_USD,columns = ['Total material cost ($)'])

    pipe_network_costs_USD = pipe_material_cost_USD.join(pipe_misc_cost_USD)
    pipe_network_costs_USD = pipe_network_costs_USD.join(pipe_labor_cost_USD)
    pipe_network_costs_USD = pipe_network_costs_USD.join(pipe_row_cost_USD)

    pipe_network_cost_total_USD = sum(pipe_total_cost_USD)

    return(pipe_network_cost_total_USD,pipe_network_costs_USD,pipe_material_cost_bymass_USD)


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt 
import os

def plot_wind_results(wind_data, site_name, latlon, results_dir, plot_wind):

    if plot_wind:
        wind_speed = [W[2] for W in wind_data]
        plt.figure(figsize=(9,6))
        plt.plot(wind_speed)
        plt.title('Wind Speed (m/s) for selected location \n {} \n lat, lon: {} \n Average Wind Speed (m/s) {}'.format(site_name,latlon,np.average(wind_speed)))
        plt.savefig(os.path.join(results_dir,'Average Wind Speed_{}'.format(site_name)),bbox_inches='tight')

def plot_pie(site_df, site_name, turbine_name, results_dir):

    group_names=['BOS', 'Soft', 'Turbine']
    group_size=[site_df['BOS'],site_df['Soft'],site_df['Turbine CapEx']]
    subgroup_names=['Array System','Export System','Substructure','Mooring System',
                'Offshore Substation','Scour Protection','Array System Installation',
                'Export System Installation', 'Offshore Substation Installation',
                'Scour Protection Installation', 'Substructure Installation','Turbine Installation',
                'Mooring System Installation',
                'construction_insurance_capex', 'decomissioning_costs',
                'construction_financing', 'procurement_contingency_costs',
                'install_contingency_costs', 'project_completion_capex',
                    'Turbine CapEx']

    subgroup_vals = site_df[subgroup_names]
    subgroup_size=[x for x in subgroup_vals]

    bos_names = ['BOS: '+name for name in subgroup_names[:13]]
    soft_names = ['Soft Cost: ' +name for name in subgroup_names[13:-1]]
    turbine_names = ['Turbine: '+name for name in subgroup_names[19:]]
    all_names = (bos_names + soft_names + turbine_names)
    subgroup_names_legs=[x for x in all_names]

    # Create colors
    a, b, c=[plt.cm.Blues, plt.cm.Reds, plt.cm.Greens]

    # First Ring (Inside)
    fig, ax = plt.subplots(figsize=(12,12))
    ax.axis('equal')
    mypie, _ = ax.pie(group_size, radius=1, labels=group_names, labeldistance=.6, colors= 
    [a(0.6), b(0.6), c(0.6)] )
    plt.setp( mypie, width=0.3, edgecolor='white')

    # Second Ring (Outside)
    mypie2, _ = ax.pie(subgroup_size, radius=1.0+0.4, 
    labels=subgroup_names, labeldistance=1, rotatelabels=True, colors=[a(0.1), a(0.2), 
    a(0.3), a(0.4), a(0.5), a(0.6), a(0.7), a(0.8), a(0.9), a(1), a(1.1), a(1.2),
                                                    a(1.3),b(.16),b(.32),b(.48),b(.64),b(.8),b(.9),c(.4)])
    plt.setp( mypie2, width=0.4, edgecolor='white')
    plt.margins(0,0)

    plt.legend(loc=(0.9, 0.1))
    handles, labels = ax.get_legend_handles_labels()

    ax.legend(handles[3:], subgroup_names_legs, loc=(0.4, 0.3))

    plt.legend(subgroup_names_legs,loc='best')
    # plt.title('ORBIT Cost Contributions for {}'.format(site_name))
    print('ORBIT Cost Contributions for {}'.format(site_name))
    plt.savefig(os.path.join(results_dir,'BOS Cost Figure {}_{}.jpg'.format(site_name,turbine_name)),bbox_inches='tight')
    # plt.show()

def plot_HOPP(combined_hybrid_power_production_hopp,
              energy_shortfall_hopp,
              combined_hybrid_curtailment_hopp,
              load,
              results_dir,
              site_name,
              atb_year,
              turbine_model,
              hybrid_plant,
              plot_power_production):

    if plot_power_production:
        plt.figure(figsize=(4,4))
        plt.title("HOPP power production")
        plt.plot(combined_hybrid_power_production_hopp[200:300],label="wind + pv")
        plt.plot(energy_shortfall_hopp[200:300],label="shortfall")
        plt.plot(combined_hybrid_curtailment_hopp[200:300],label="curtailment")
        plt.plot(load[200:300],label="electrolyzer rating")
        plt.xlabel("Time (hour)")
        plt.ylabel("Power Production (kW)")
        # plt.ylim(0,250000)
        plt.legend()
        plt.tight_layout()
        plt.savefig(os.path.join(results_dir,'HOPP Power Production_{}_{}_{}'.format(site_name,atb_year,turbine_model)),bbox_inches='tight')
        # plt.show()

    print("Turbine Power Output (to identify powercurve impact): {0:,.0f} kW".format(hybrid_plant.wind.annual_energy_kw))
    print("Wind Plant CF: {}".format(hybrid_plant.wind.capacity_factor))
    print('LCOE: ', hybrid_plant.lcoe_real.hybrid)
    # print("LCOE: {}"].format(hybrid_plant.lcoe_real.hybrid))

def plot_battery_results(combined_hybrid_curtailment_hopp, 
                         energy_shortfall_hopp,
                         combined_pv_wind_storage_power_production_hopp,
                         combined_hybrid_power_production_hopp,
                         battery_SOC,
                         battery_used,
                         results_dir,
                         site_name,atb_year,turbine_model,
                         load,
                         plot_battery):

    if plot_battery:
        plt.figure(figsize=(9,6))
        plt.subplot(311)
        plt.plot(combined_hybrid_curtailment_hopp[200:300],label="curtailment")
        plt.plot(energy_shortfall_hopp[200:300],label="shortfall")
        plt.title('Energy Curtailment and Shortfall')
        plt.legend()
        

        plt.subplot(312)
        plt.plot(combined_pv_wind_storage_power_production_hopp[200:300],label="wind+pv+storage")
        plt.plot(combined_hybrid_power_production_hopp[200:300],"--",label="wind+pv")
        plt.plot(load[200:300],"--",label="electrolyzer rating")
        plt.legend()
        plt.title("Hybrid Plant Power Flows with and without storage")
        plt.tight_layout()
        
        plt.subplot(313)
        plt.plot(battery_SOC[200:300],label="state of charge")
        plt.plot(battery_used[200:300],"--",label="battery used")
        plt.title('Battery State')
        plt.legend()
        plt.savefig(os.path.join(results_dir,'HOPP Full Power Flows_{}_{}_{}'.format(site_name,atb_year,turbine_model)),bbox_inches='tight')
        # plt.show()

def plot_h2_results(H2_Results, 
                    electrical_generation_timeseries,
                    results_dir,
                    site_name,atb_year,turbine_model,
                    load,
                    plot_h2):

    if plot_h2:
        hydrogen_hourly_production = H2_Results['hydrogen_hourly_production']
        plt.figure(figsize=(8,8))
        plt.subplot(411)
        plt.plot(electrical_generation_timeseries[200:300])
        plt.ylim(0,max(electrical_generation_timeseries[200:300])*1.2)
        plt.plot(load[200:300],label="electrolyzer rating")
        plt.legend()
        plt.title("Energy to electrolyzer (kW)")

        plt.subplot(412)
        plt.plot(hydrogen_hourly_production[200:300])
        plt.ylim(0,max(hydrogen_hourly_production[200:300])*1.2)
        plt.title("Hydrogen production rate (kg/hr)")

        
        plt.subplot(413)
        plt.plot(H2_Results['electrolyzer_total_efficiency'][200:300])
        plt.ylim(0,1)
        plt.title("Electrolyzer Total Efficiency (%)")
        
        
        plt.subplot(414)
        plt.plot(H2_Results['water_hourly_usage'][200:300],"--",label="Hourly Water Usage")
        plt.legend()
        plt.title('Hourly Water Usage (kg/hr) \n' 'Total Annual Water Usage: {0:,.0f}kg'.format(H2_Results['water_annual_usage']))
        plt.tight_layout()
        plt.xlabel('Time (hours)')
        plt.savefig(os.path.join(results_dir,'Electrolyzer Flows_{}_{}_{}'.format(site_name,atb_year,turbine_model)),bbox_inches='tight')
        # plt.show()

def plot_desal_results(fresh_water_flowrate,
                        feed_water_flowrate,
                        operational_flags,
                        results_dir,
                        site_name,atb_year,turbine_model,
                        plot_desal):

    if plot_desal:
        plt.figure(figsize=(10,5))
        plt.subplot(1,2,1)
        plt.plot(fresh_water_flowrate[200:300],"--",label="Freshwater flowrate from desal")
        plt.plot(feed_water_flowrate[200:300],"--",label="Feedwater flowrate to desal")
        plt.legend()
        plt.title('Freshwater flowrate (m^3/hr) from desal  (Snapshot)')
        # plt.show()

        plt.subplot(1,2,2)
        plt.plot(operational_flags[200:300],"--",label="Operational Flag")
        plt.legend()
        plt.title('Desal Equipment Operational Status (Snapshot) \n 0 = Not enough power to operate \n 1 = Operating at reduced capacity \n 2 = Operating at full capacity')
        plt.savefig(os.path.join(results_dir,'Desal Flows_{}_{}_{}'.format(site_name,atb_year,turbine_model)),bbox_inches='tight')
        # plt.show()

def plot_hvdcpipe(total_export_system_cost,
                  total_h2export_system_cost,
                  site_name,
                  atb_year,
                  dist_to_port_value,
                  results_dir):

    # create data
    if plot_hvdcpipe:
        barx = ['HVDC', 'Pipeline']
        #cost_comparison_hvdc_pipeline = [capex_pipeline,total_export_system_cost]
        cost_comparison_hvdc_pipeline = [total_export_system_cost, total_h2export_system_cost]
        plt.figure(figsize=(9,6))
        plt.bar(barx, cost_comparison_hvdc_pipeline)

        plt.ylabel("$USD")
        plt.legend(["Total CAPEX"])
        #plt.title("H2 Pipeline vs HVDC cost\n {}\n Model:{}".format(site_name,in_dict['pipeline_model']))
        plt.title("H2 Pipeline vs HVDC cost\n {}\n Model: ASME Pipeline".format(site_name))
        plt.savefig(os.path.join(results_dir,'Pipeline Vs HVDC Cost_{}_{}_{}'.format(site_name,atb_year,dist_to_port_value)))
        #plt.show()
